# Author: Paul Kienzle
"""
Reflectometry models

Reflectometry models consist of 1-D stacks of layers. Layers are joined
by gaussian interfaces. The layers themselves may be uniform, or the
scattering density may vary with depth in the layer.

.. Note::
   By importing model, the definition of :class:`material.Scatterer <refl1d.sample.material.Scatterer>`
   changes so that materials can be stacked into layers using operator
   overloading:
   - the | operator, (previously known as "bitwise or") joins stacks
   - the * operator repeats stacks (n times, n is an int)

   This will affect all instances of the Scatterer class, and all of its subclasses.
"""

# TODO: xray has smaller beam spot
# => smaller roughness
# => thickness depends on where the beam spot hits the sample
# Xray thickness variance = neutron roughness - xray roughness

__all__ = ["Repeat", "Slab", "Stack", "Layer"]

from copy import copy
from dataclasses import dataclass, field
from typing import List, Literal, Optional, Union

import numpy as np
from bumps.parameter import Calculation, Parameter

from .. import profile
from ..probe.probe import NeutronProbe, XrayProbe
from ..sample import material as mat
from .magnetism import BaseMagnetism


# @dataclass(init=False)
class Layer:  # Abstract base class
    """
    Component of a material description.

    thickness (Parameter: angstrom)
        Thickness of the layer
    interface (Parameter: angstrom)
        Interface for the top of the layer.
    magnetism (Magnetism info)
        Magnetic profile anchored to the layer.
    """

    name: str
    thickness: Parameter
    interface: Optional[Parameter] = None
    magnetism: Optional[BaseMagnetism] = None

    # Trap calls to set magnetism attr so we can update the magnetism parameter
    # names with the layer name when we assign magnetism to the layer
    def __setattr__(self, name, value):
        if name == "magnetism" and value is not None:
            value.set_layer_name(str(self))
        object.__setattr__(self, name, value)

    @property
    def ismagnetic(self):
        return self.magnetism is not None

    def constraints(self):
        """
        Constraints
        """
        return self.thickness >= 0, self.interface >= 0

    def find(self, z):
        """
        Find the layer at depth z.

        Returns layer, start, end
        """
        return self, 0, self.thickness.value

    def parameters(self):
        """
        Returns a dictionary of parameters specific to the layer.  These will
        be added to the dictionary containing interface, thickness and magnetism
        parameters.
        """

    def layer_parameters(self):
        pars = {"thickness": self.thickness}
        if self.interface is not None:
            pars["interface"] = self.interface
        if self.magnetism is not None:
            pars["magnetism"] = self.magnetism.parameters()
        pars.update(self.parameters())
        return pars

    def render(self, probe, slabs):
        """
        Use the probe to render the layer into a microslab representation.
        """

    def penalty(self):
        """
        Return a penalty value associated with the layer.  This should be
        zero if the parameters are valid, and increasing as the parameters
        become more invalid.  For example, if total volume fraction exceeds
        unity, then the penalty would be the amount by which it exceeds
        unity, or if z values must be sorted, then penalty would be the
        amount by which they are unsorted.

        Note that penalties are handled separately from any probability of
        seeing a combination of layer parameters; the final solution to the
        problem should not include any penalized points.
        """
        return 0

    def __str__(self):
        """
        Print shows the layer name
        """
        return getattr(self, "name", repr(self))

    # Define a little algebra for composing samples
    # Layers can be stacked, repeated, or have length/roughness/magnetism set
    def __or__(self, other):
        """Join two layers to make a stack"""
        stack = Stack()
        stack.add(self)
        stack.add(other)
        return stack

    def __mul__(self, other):
        """Repeat a stack or complex layer"""
        if not isinstance(other, int) or not other > 1:
            raise TypeError("Repeat count must be an integer > 1")
        if isinstance(self, Slab):
            raise TypeError("Cannot repeat single slab")
        stack = Stack()
        stack.add(self)
        r = Repeat(stack=stack, repeat=other)
        return r

    def __rmul__(self, other):
        return self.__mul__(other)

    def __call__(self, thickness=None, interface=None, magnetism=None):
        c = copy(self)
        # Only set values if they are not None so that defaults
        # carry over from the copied layer
        if thickness is not None:
            c.thickness = Parameter.default(thickness, limits=(0, None), name=self.name + " thickness")
        if interface is not None:
            c.interface = Parameter.default(interface, limits=(0, None), name=self.name + " interface")
        if magnetism is not None:
            c.magnetism = magnetism
        return c


def _parinit(p, v):
    """
    If v is a parameter use v, otherwise use p but with value v.
    """
    if isinstance(v, Parameter):
        p = v
    else:
        p.set(v)
    return p


def _parcopy(p, v):
    """
    If v is a parameter use v, otherwise use a copy of p but with value v.
    """
    if isinstance(v, Parameter):
        p = v
    else:
        p = copy(p)
        p.set(v)
    return p


@dataclass(init=False)
class Slab(Layer):
    """
    A block of material.
    """

    name: str
    thickness: Parameter
    interface: Optional[Parameter] = None
    magnetism: Optional[BaseMagnetism]
    material: Union[mat.SLD, mat.MaterialTypes, mat.Vacuum]

    def __init__(self, material=None, thickness=0, interface=0, name=None, magnetism=None):
        if name is None:
            name = material.name
        self.name = name
        self.material = material
        self.thickness = Parameter.default(thickness, limits=(0, None), name=name + " thickness")
        self.interface = Parameter.default(interface, limits=(0, None), name=name + " interface")
        self.magnetism = magnetism

    def parameters(self):
        return {"material": self.material.parameters()}

    def render(self, probe, slabs):
        rho, irho = self.material.sld(probe)
        w = self.thickness.value
        sigma = self.interface.value
        # print "rho", rho
        # print "irho", irho
        # print "w", w
        # print "sigma", sigma
        slabs.append(rho=rho, irho=irho, w=w, sigma=sigma)

    def __str__(self):
        return self.name
        # return str(self.material)

    def __repr__(self):
        return "Slab(" + repr(self.material) + ")"


@dataclass(init=False)
class Stack(Layer):
    # TODO: Finish this docstring
    """
    Reflectometry layer stack

    A reflectometry sample is defined by a stack of layers. Each layer
    has an interface describing how the top of the layer interacts with
    the bottom of the overlaying layer. The stack may contain
    """

    name: str
    interface: Literal[None]
    layers: List[Union["Slab", "Repeat"]]
    thickness: Parameter = field(metadata={"description": "always equals the sum of the layer thicknesses"})

    def __init__(
        self,
        layers: Optional[Union["Stack", List[Union["Slab", "Repeat"]]]] = None,
        name: str = "Stack",
        interface=None,
        thickness: Optional[Union[Parameter, float]] = None,
    ):
        self.name = name
        self.interface = None
        self._layers = []
        # make sure thickness.id is persistent through serialization:
        thickness = thickness if thickness is not None else 0
        self.thickness = Parameter.default(thickness, name=name + " thickness")
        self._set_thickness()
        if layers is not None:
            self.add(layers)

    @property
    def layers(self):
        return self._layers

    @property
    def ismagnetic(self):
        return any(p.ismagnetic for p in self._layers)

    def find(self, z):
        """
        Find the layer at depth z.

        Returns layer, start, end
        """
        offset = 0
        for L in self._layers:
            dz = L.thickness.value
            if z < offset + dz:
                break
            offset += dz
        else:
            L = self._layers[-1]
            offset -= dz

        L, start, end = L.find(z - offset)
        return L, start + offset, end + offset

    def add(self, other):
        if isinstance(other, Stack):
            self._layers.extend(other._layers)
        elif isinstance(other, Repeat):
            self._layers.append(other)
        else:
            try:
                L = iter(other)
            except TypeError:
                L = [other]
            self._layers.extend(_check_layer(el) for el in L)

    def __setstate__(self, state):
        # this is a temporary shim (2024-10-29), to accomodate objects that were pickled
        # before the custom __getstate__ was removed.
        # TODO: the entire __setstate__ can be removed someday (e.g. in 2026?)
        if isinstance(state, tuple):
            self.interface, self._layers, self.name, self.thickness = state
        else:
            self.__dict__.update(state)

    def __copy__(self):
        stack = Stack()
        stack.interface = self.interface
        stack._layers = self._layers[:]
        stack.thickness = self.thickness
        return stack

    def __len__(self):
        return len(self._layers)

    def __str__(self):
        return " | ".join("%s(%.3g)" % (L, L.thickness.value) for L in self._layers)

    def __repr__(self):
        return "Stack(" + ", ".join(repr(L) for L in self._layers) + ")"

    def parameters(self):
        layers = [L.layer_parameters() for L in self._layers]
        return {"thickness": self.thickness, "layers": layers}

    def penalty(self):
        return sum(L.penalty() for L in self._layers)

    def _calculate_thickness(self):
        return sum(L.thickness.value for L in self._layers)

    def _set_thickness(self):
        desc = """sum of the thickness of the component layers in the stack"""
        calculation = Calculation(description=desc)
        calculation.set_function(self._calculate_thickness)
        self.thickness.slot = calculation

    def render(self, probe, slabs):
        """
        Render the stack into slabs.
        """
        if any(layer.magnetism is not None for layer in self._layers):
            return self._render_magnetic(probe, slabs)
        else:
            return self._render_nonmagnetic(probe, slabs)

    def _render_nonmagnetic(self, probe, slabs):
        """
        Render and sld stack in which no layers are magnetic.
        """
        for layer in self._layers:
            layer.render(probe, slabs)

    def _render_magnetic(self, probe, slabs):
        """
        Render and sld stack in which some layers are magnetic.

        If the magnetism interface above or below is left unspecified, the
        corresponding nuclear interface is used.
        """
        magnetism = None
        end_layer = -1
        for i, layer in enumerate(self._layers):
            # Trigger start of a magnetic layer
            if layer.magnetism is not None:
                if magnetism is not None:
                    raise IndexError("magnetic layer %s overlap" % magnetism)
                magnetism = layer.magnetism
                # import sys; print >>sys.stderr, "magnetism", magnetism
                anchor = slabs.thickness() + magnetism.dead_below.value
                s_below = (
                    np.nan
                    if i == 0
                    else magnetism.interface_below.value
                    if magnetism.interface_below is not None
                    else slabs.surface_sigma
                )
                end_layer = i + magnetism.extent - 1

            # Render nuclear layer
            layer.render(probe, slabs)

            # Wait for end of magnetic layer
            if i == end_layer:
                s_above = (
                    magnetism.interface_above.value if magnetism.interface_above is not None else slabs.surface_sigma
                )
                w = (slabs.thickness() - magnetism.dead_above.value) - anchor
                magnetism.render(probe, slabs, thickness=w, anchor=anchor, sigma=(s_below, s_above))
                magnetism = None

        if magnetism:
            raise IndexError("magnetic layer %s is incomplete" % magnetism)

    def _plot(self, dz=1, roughness_limit=0):
        # TODO: unused?
        import matplotlib.pyplot as plt

        from . import material

        neutron_probe = NeutronProbe(T=np.arange(0, 5, 100), L=5.0)
        xray_probe = XrayProbe(T=np.arange(0, 5, 100), L=1.54)
        slabs = profile.Microslabs(1, dz=dz)

        plt.subplot(211)
        cache = material.ProbeCache(xray_probe)
        slabs.clear()
        self.render(cache, slabs)
        z, rho, irho = slabs.step_profile()
        plt.plot(z, rho, "-g", z, irho, "-b")
        z, rho, irho = slabs.smooth_profile(dz=1, roughness_limit=roughness_limit)
        plt.plot(z, rho, ":g", z, irho, ":b")
        plt.legend(["rho", "irho"])
        plt.xlabel("depth (A)")
        plt.ylabel("SLD (10^6 inv A**2)")
        plt.text(0.05, 0.95, r"Cu-$K_\alpha$ X-ray", va="top", ha="left", transform=plt.gca().transAxes)

        plt.subplot(212)
        cache = material.ProbeCache(neutron_probe)
        slabs.clear()
        self.render(cache, slabs)
        z, rho, irho = slabs.step_profile()
        plt.plot(z, rho, "-g", z, irho, "-b")
        z, rho, irho = slabs.smooth_profile(dz=1, roughness_limit=roughness_limit)
        plt.plot(z, rho, ":g", z, irho, ":b")
        plt.legend(["rho", "irho"])
        plt.xlabel("depth (A)")
        plt.ylabel("SLD (10^6 inv A**2)")
        plt.text(0.05, 0.95, "5 A neutron", va="top", ha="left", transform=plt.gca().transAxes)

    # Stacks as lists
    def _find_by_material(self, target):
        """
        Iterate over all layers that have the given material.
        """
        for i, layer in enumerate(self._layers):
            if hasattr(layer, "stack"):
                for sub in layer.stack._find_by_material(target):
                    yield sub
            elif hasattr(layer, "material"):
                if id(layer.material) == id(target):
                    yield self, i

    def _find_by_name(self, target):
        """
        Iterate over all layers that have the given name.
        """
        for i, layer in enumerate(self._layers):
            if hasattr(layer, "stack"):
                for sub in layer.stack._find_by_name(target):
                    yield sub
            else:
                if str(layer) == target:
                    yield self, i

    def _lookup(self, idx):
        """
        Lookup a layer by integer index, name, material or (material, repeat) if not the first
        occurrence of the material in the sample.  Search is depth first.  Returns the stack
        or substack that contains the material, and the index in that stack.
        """
        if isinstance(idx, int):
            return self, idx

        if isinstance(idx, slice):
            start = (self, 0) if idx.start is None else self._lookup(idx.start)
            stop = (self, len(self)) if idx.stop is None else self._lookup(idx.stop)
            if start[0] != stop[0]:
                raise IndexError("start and and stop of sample slice must be in the same stack")
            return start[0], slice(start[1], stop[1], idx.step)

        # Check for lookup of the nth occurrence of a given layer
        if isinstance(idx, tuple):
            target, count = idx
        else:
            target, count = idx, 1

        # Check if lookup by material or by name
        if isinstance(target, mat.Scatterer):
            sequence = self._find_by_material(target)
        elif isinstance(target, str):
            sequence = self._find_by_name(target)
        else:
            raise TypeError("expected integer, material or layer name as sample index")

        # Move to the nth item in the sequence
        i = -1
        for i, el in enumerate(sequence):
            if i + 1 == count:
                return el
        if i == -1:
            raise IndexError("layer %s not found" % str(target))
        else:
            raise IndexError("only found %d layers of %s" % (str(target), i + 1))

    def __getitem__(self, idx):
        # import sys;print >>sys.stderr, "lookup idx", idx
        stack, idx = self._lookup(idx)
        # print >>sys.stderr, "found", idx
        if isinstance(idx, slice):
            newstack = Stack()
            newstack._layers = stack._layers[idx]
            return newstack
        else:
            return stack._layers[idx]

    def __setitem__(self, idx, other):
        stack, idx = self._lookup(idx)
        if isinstance(idx, slice):
            if isinstance(other, Stack):
                stack._layers[idx] = other._layers
            else:
                stack._layers[idx] = [_check_layer(el) for el in other]
        else:
            stack._layers[idx] = _check_layer(other)

    def __delitem__(self, idx):
        stack, idx = self._lookup(idx)
        # works the same for slices and individual indices
        del stack._layers[idx]

    def insert(self, idx, other):
        """
        Insert structure into a stack.  If the inserted element is
        another stack, the stack will be expanded to accommodate.  You
        cannot make nested stacks.
        """
        stack, idx = self._lookup(idx)
        if isinstance(other, Stack):
            for i, L in enumerate(other._layers):
                stack._layers.insert(idx + i, L)
        elif isinstance(other, Repeat):
            stack._layers.insert(idx, other)
        else:
            try:
                other = iter(other)
            except Exception:
                other = [other]
            for i, L in enumerate(other):
                stack._layers.insert(idx + i, _check_layer(L))

    # Define a little algebra for composing samples
    # Stacks can be repeated or extended
    def __mul__(self, other):
        if isinstance(other, Parameter):
            pass
        elif isinstance(other, int) and other > 1:
            pass
        else:
            raise TypeError("Repeat count must be an integer > 1")
        s = Repeat(stack=self, repeat=other)
        return s

    def __rmul__(self, other):
        return self.__mul__(other)

    def __or__(self, other):
        stack = Stack()
        stack.add(self)
        stack.add(other)
        return stack

    render.__doc__ = Layer.render.__doc__


def _check_layer(el):
    if isinstance(el, Layer):
        return el
    elif isinstance(el, mat.Scatterer):
        return Slab(el)
    else:
        raise TypeError(f"Can only stack materials and layers, not {el} (of type {type(el)})")


@dataclass(init=False)
class Repeat(Layer):
    """
    Repeat a layer or stack.

    If an interface parameter is provide, the roughness between the
    multilayers may be different from the roughness between the repeated
    stack and the following layer.

    Note: Repeat is not a type of Stack, but it does have a stack inside.
    """

    name: str
    interface: Parameter
    magnetism: Optional[BaseMagnetism]
    repeat: Parameter
    stack: Stack
    thickness: Parameter

    def __init__(
        self, stack, repeat=1, interface=None, name=None, magnetism=None, thickness: Optional[Parameter] = None
    ):
        name = "multilayer" if name is None else name
        interface = stack[-1].interface.value if interface is None else interface
        self.magnetism = magnetism
        self.name = name
        self.repeat = Parameter(repeat, limits=(0, None), discrete=True, name=name + " repeats")
        self.stack = stack
        # intentionally ignoring what was in thickness argument...
        # is this a good idea?
        self.thickness = Parameter(name=name + " thickness")
        self.thickness.equals(self.stack.thickness * self.repeat)
        self.interface = Parameter.default(interface, limits=(0, None), name=name + " top interface")

    def __setstate__(self, state):
        # this is a temporary shim (2024-10-29), to accomodate objects that were pickled
        # before the custom __getstate__ was removed.
        # TODO: the entire __setstate__ can be removed someday (e.g. in 2026?)
        if isinstance(state, tuple):
            self.interface, self.repeat, self.name, self.stack = state
        else:
            self.__dict__.update(state)

    def penalty(self):
        return self.stack.penalty()

    @property
    def ismagnetic(self):
        return self.magnetism is not None or self.stack.ismagnetic

    def find(self, z):
        """
        Find the layer at depth z.

        Returns layer, start, end
        """
        n = self.repeat.value
        unit = self.thickness.value
        if z < n * unit:
            offset = int(z / unit) * unit
            L, start, end = self.stack.find(z - offset)
            return L, start + offset, end + offset
        else:
            offset = n * unit
            L, start, end = self.stack.find(unit)
            return L, start + offset, end + offset

    # Stacks as lists
    def __getitem__(self, idx):
        return self.stack[idx]

    def __setitem__(self, idx, other):
        self.stack[idx] = other

    def __delitem__(self, idx):
        del self.stack[idx]

    def parameters(self):
        pars = {
            "stack": self.stack.parameters(),
            "repeat": self.repeat,
            "thickness": self.thickness,
            "interface": self.interface,
        }
        if self.magnetism:
            pars["magnetism"] = self.magnetism.parameters()
        return pars

    def render(self, probe, slabs):
        nr = self.repeat.value
        if nr <= 0:
            return
        mark = len(slabs)
        self.stack.render(probe, slabs)
        slabs.repeat(mark, int(nr), interface=self.interface.value)

    def __str__(self):
        return "(%s)x%d" % (str(self.stack), self.repeat.value)

    def __repr__(self):
        return "Repeat(%s, %d)" % (repr(self.stack), self.repeat.value)


# Extend the material.Scatterer class so that any scatter can be
# implicitly turned into a slab.
def _material_stacker():
    """
    Allow materials to be used in the stack algebra.  Material can be
    called with thickness, interface, magnetism and turned into a slab.
    So instead of::

        sample = Slab(Si, 0, 5) | Slab(Ni, 100, 5) | Slab(air)

    models can use::

        sample = Si(0, 5) | Ni(100, 5) | air

    WARNING: this adds __or__ and __call__ methods to the material.Scatterer
    base class.  This is a nasty thing to do since those who have to debug
    the Scatterer later will not know where to look for these class
    attributes.  On the flip side, changing the base class definition saves
    us from the nastier problem of having to create a sister hierarchy of
    stackable scatterers mirroring the scatterers hierarchy, or the nasty
    problem of a circular definition that Slab depends on Scatterer and
    Scatterer depends on Slab.
    """

    # Note: should have been add these as a Mixin class, but couldn't
    # get it to work on python 3.3
    def __or__(self, other):
        # need __or__ for stacks which start with a bare material, such
        # as Si | air
        stack = Stack()
        # Note: stack.add() converts materials to slabs
        stack.add(self)
        stack.add(other)
        return stack

    def __call__(self, thickness=0, interface=0, magnetism=None):
        slab = Slab(material=self, thickness=thickness, interface=interface, magnetism=magnetism)
        return slab

    mat.Scatterer.__or__ = __or__
    mat.Scatterer.__call__ = __call__


_material_stacker()



"""
Reflectivity plugin for fitting GUI.

Note that the fitting infrastructure is still heavily tied to the reflectivity
modeling program, and this represents only the first tiny steps to separating
the two.
"""

__all__ = ["data_view", "model_view", "new_model", "calc_errors", "show_errors"]

from typing import cast
import numpy as np

from .migrations import migrate
from bumps.fitproblem import FitProblem
from ..experiment import Experiment
from ..sample.materialdb import air, silicon
from ..probe.data_loaders import ncnrdata as NCNR
from ..uncertainty import calc_errors, show_errors

# List of modules that contain dataclasses for the saved json file format


# These are names used by the driver
def data_view():
    from refl1d.wx_gui.data_view import DataView

    return DataView


def model_view():
    from refl1d.wx_gui.model_view import ModelView

    return ModelView


def load_model(filename):
    # TODO: bumps plugin api needs to allow options for loader
    options = None
    if filename.endswith(".staj") or filename.endswith(".sta"):
        from ..probe.data_loaders.stajconvert import load_mlayer

        return FitProblem[Experiment](load_mlayer(filename))
        # fit_all(problem.fitness, pmp=20)
    elif filename.endswith(".zip"):
        from bumps.fitproblem import load_problem

        # Note: bumps.vfs.vfs_init() must be called very early
        try:
            from bumps.vfs import ZipFS
        except ImportError:
            raise NotImplementedError("need newer bumps to load model from zip")
        with ZipFS(filename) as zf:
            for f in zf.filelist:
                if f.filename.endswith(".py"):
                    return cast(FitProblem[Experiment], load_problem(f.filename, options=options))
    elif filename.endswith(".json"):
        from bumps.serialize import load_file

        return cast(FitProblem[Experiment], load_file(filename))
    else:
        return None


def save_json(problem, basename):
    from bumps.serialize import save

    json_filename = basename + "-expt.json"
    save(json_filename, problem)


def new_model():
    stack = silicon(0, 10) | air
    instrument = NCNR.NG1()
    probe = instrument.probe(T=np.linspace(0, 5, 200), Tlo=0.2, slits_at_Tlo=2)
    M = Experiment(sample=stack, probe=probe)
    problem = FitProblem(M)
    return problem


def migrate_serialized(model_dict):
    _, migrated = migrate(model_dict)
    return migrated



"""
Support files for the application.

This includes tools to help with testing, documentation, command line
parsing, etc. which are specific to this application, rather than general
utilities.
"""

import os
from pathlib import Path


def get_data_path():
    """
    Locate the examples directory.
    """
    # Check for data path in the environment
    key = "REFL1D_DATA"
    if key in os.environ:
        path = os.environ[key]
        if not os.path.isdir(path):
            raise RuntimeError("Path in environment %s not a directory" % key)
        return path

    # Check for data next to the package.
    try:
        root = Path(__file__).resolve().parent.parent.parent
        return root / "doc" / "examples"
    except Exception:
        raise RuntimeError("Could not find sample data")


_REGISTRY = {
    # List the example datasets that are needed for demos and doctests.
    # Each line looks like:
    #
    #     'dataset': ['subdirectory', 'filename'],
    #
    "spin_valve01.refl": ["spinvalve", "spin_valve01.refl"],
    "chale207.refl": ["polymer", "10ndt001.refl"],
    "10ndt001.refl": ["polymer", "10ndt001.refl"],
    "lha03_255G.refl": ["spinvalve", "n101Gc1.refl"],
}


def sample_data(file):
    examples = get_data_path()
    if file in _REGISTRY:
        return os.path.join(examples, *_REGISTRY[file])
    else:
        raise ValueError("Sample dataset %s not available" % file)



# -*- coding: utf-8 -*-
"""
Visual representation of model uncertainty.

For reflectivity models, this aligns and plots a set of profiles chosen
from the parameter uncertainty distribution, and plots the distribution
of the residual values.

Use *run_errors* in a model file to reload the results of a batch DREAM fit.
"""

__all__ = [
    "align_profiles",
    "calc_errors",
    "reload_errors",
    "run_errors",
    "show_errors",
    "show_profiles",
    "show_residuals",
]

import os
import sys

import numpy as np
from bumps.errplot import reload_errors
from bumps.plotutil import dhsv, form_quantiles, next_color, plot_quantiles

from .sample.reflectivity import BASE_GUIDE_ANGLE
from .utils import asbytes

# CONTOURS = (68, 95, 100)
# CONTOURS = (57, 68, 84, 95, 100)
CONTOURS = (68, 95)

# TODO: we should just keep a certain number of evaluations as a matter of
#       course during sampling rather than recomputing them after the fact.
# TODO: want similar code for covariance matrix based forward analysis
# TODO: need to delegate accumulation of models and plotting to Fitness
# TODO: move run_errors to align.py and use argparse to process sys.argv
# TODO: add controls for the additional parameters to the command line


def run_errors(**kw):
    """
    Command line tool for generating error plots from models.

    Type the following to regenerate the profile contour plots plots:

        $ refl1d align <model>.py <store> [<align>] [0|1|2|n]

    Align is either auto for the current behaviour, or it is an interface
    number. You can align on the center of a layer by adding 0.5 to the
    interface number. You can count interfaces from the surface by prefixing
    with R.  For example, 0 is the substrate interface, R1 is the surface
    interface, 2.5 is the the middle of layer 2 above the substrate.

    You can plot the profiles and residuals on one plot by setting plots to 1,
    on two separate plots by setting plots to 2, or each curve on its own
    plot by setting plots to n. Plots are saved in <store>/<model>-err#.png.
    If plots is 0, then no plots are created.

    Additional parameters include:

        *nshown*, *random* :

            see :func:`bumps.errplot.calc_errors_from_state`

        *contours*, *npoints*, *plots*, *save* :

            see :func:`show_errors`
    """

    load = {"model": None, "store": None, "nshown": 50, "random": True}
    show = {"align": "auto", "plots": 2, "contours": CONTOURS, "npoints": 400, "save": None}

    for k, v in kw.items():
        if k in load:
            load[k] = v
        elif k in show:
            show[k] = v
        else:
            raise TypeError("Unknown argument " + k)

    if len(sys.argv) > 2:
        load["model"], load["store"] = sys.argv[1], sys.argv[2]

        align_str = sys.argv[3] if len(sys.argv) >= 4 else "0"
        if align_str[0] == "R":
            align_str = "-" + align_str[1:]
        show["align"] = float(align_str) if align_str != "auto" else align_str
        plots_str = sys.argv[4] if len(sys.argv) >= 5 else "2"
        show["plots"] = int(plots_str) if plots_str != "n" else plots_str
        # print align, align_str, len(sys.argv), sys.argv

    if not load["store"] or not load["model"]:
        _usage()
        sys.exit(0)

    if show["save"] is None:
        name, _ = os.path.splitext(os.path.basename(load["model"]))
        show["save"] = os.path.join(load["store"], name)

    print("loading... this may take awhile")
    errors = reload_errors(**load)
    print("showing...")
    show_errors(errors, **show)

    if show["plots"] != 0:
        import matplotlib.pyplot as plt

        plt.show()
    sys.exit(0)  # Force refl1d to terminate.


def _usage():
    print(run_errors.__doc__)


def calc_errors(problem, points):
    """
    Align the sample profiles and compute the residual difference from the
    measured reflectivity for a set of points.

    The points should be sampled from the posterior probability
    distribution computed from MCMC, bootstrapping or sampled from
    the error ellipse calculated at the minimum.

    Each of the returned arguments is a dictionary mapping model number to
    error sample data as follows:

    Returns (profiles, slabs, Q, residuals).

    *profiles*

        Arrays of (z, rho, irho) for non-magnetic models or arrays
        of (z, rho, irho, rhoM, thetaM) for magnetic models.  There
        will be one set of arrays returned per error sample.

    *slabs*

        Array of slab thickness for the layers in the models.  There
        will be one array returned per error sample.  Using slab thickness,
        profiles can be aligned on interface boundaries and layer centers.

    *Q*

        Array of Q values for the data points in the model.  The data
        points are the same for all error samples, so only one Q array
        is needed per model.

    *residuals*

        Array of (theory-data)/uncertainty for each data point in
        the measurement.  There will be one array returned per error sample.
    """
    # Find Q
    Q = [_residQ(m) for m in _experiments(problem)]

    # Put best at slot 0, no alignment
    data = [_eval_point(problem, problem.getp())]
    for p in points:
        data.append(_eval_point(problem, p))

    profiles, slabs, residuals = zip(*data)

    # TODO: return sane datastructure
    # Make a hashable version of model which just contains the name
    # attribute, which is all that the rest of this code accesses.
    models = [_HashableModel(m, i) for i, m in enumerate(_experiments(problem))]

    profiles = {h: [v[k] for v in profiles] for k, h in enumerate(models)}
    slabs = {h: [v[k] for v in slabs] for k, h in enumerate(models)}
    residuals = {h: np.asarray([v[k] for v in residuals]).T for k, h in enumerate(models)}
    Q = {h: Q[k] for k, h in enumerate(models)}

    # from .pstruct import pstruct, sstruct
    # print("profiles", sstruct(profiles))
    # print("slabs", sstruct(slabs))
    # print("residuals", sstruct(residuals))
    # print("Q", sstruct(Q))
    # import sys; sys.exit()

    return profiles, slabs, Q, residuals


class _HashableModel:
    name: str
    index: int

    def __init__(self, model, index):
        self.name = model.name if model.name is not None else f"M{index}"
        self.index = index

    def __str__(self):
        return f"model {self.name}: {self.index}"


def _eval_point(problem, p):
    problem.chisq_str()  # Force reflectivity recalculation
    problem.setp(p)
    profiles, residuals, slabs = [], [], []
    for m in _experiments(problem):
        D = m.residuals()
        residuals.append(D + 0)
        slabs_i = [L.thickness.value for L in m.sample[1:-1]]
        slabs.append(np.array(slabs_i))
        if m.ismagnetic:
            z, rho, irho, rhoM, thetaM = m.magnetic_smooth_profile()
            profiles.append((z + 0, rho + 0, irho + 0, rhoM + 0, thetaM + 0))
        else:
            z, rho, irho = m.smooth_profile()
            profiles.append((z + 0, rho + 0, irho + 0))
    return profiles, slabs, residuals


def _experiments(problem):
    """
    Cycle through experiments yielding (k, m) pairs for each experiment.

    The iterator is necessary because bumps substitutes the values from the
    free parameters into the fitness via the model iterator in problem. In
    order to keep the parametersets in sync we need to repeat that iteration
    each time.
    """
    for m in problem.models:
        parts = getattr(m, "parts", [m])
        for p in parts:
            yield p


def _residQ(m):
    if m.probe.polarized:
        return np.hstack([xs.Q for xs in m.probe.xs if xs is not None])
    else:
        return m.probe.Q


def align_profiles(profiles, slabs, align):
    """
    Align profiles for each sample
    """
    return dict((m, _align_profile_set(profiles[m], slabs[m], align)) for m in profiles.keys())


def show_errors(errors, contours=CONTOURS, npoints=200, align="auto", plots=1, save=None, fig=None):
    """
    Plot the aligned profiles and the distribution of the residuals for
    profiles and residuals returned from calc_errors.

    *contours* can be a list of percentiles or [].  If percentiles
    are given, then show uncertainty using a contour plot with the
    given levels, otherwise just overplot sample lines.
    *contours* defaults to [68, 95, 100].

    *npoints* is the number of points to use when generating the
    profile contour.  Since the z values for the various lines do not
    correspond, the contour generator interpolates the entire profile
    range with linear spacing using this number of points.

    *align* is the interface number plus fractional distance within
    the layer following the interface.  For example, use 0 for the
    substrate interface, use -1 for the surface interface, or use 2.5
    for the center of the second slab above the substrate. If *align='auto'*
    then choose an offset that minimizes the cross-correlation between the
    first profile and the current profile.

    *plots* is the number of plots to use (1, 2, or 'n').

    *save* is the basename of the plot to save.  This should usually
    be "<store>/<model>".  The program will add '-err#.png' where '#'
    is the number of the plot.
    """
    import matplotlib.pyplot as plt

    if fig is not None and plots != 1:
        raise ValueError("can only pass in a figure object if exactly 1 plot is requested")

    if plots == 0:  # Don't create plots, just save the data
        _save_profile_data(errors, contours=contours, npoints=npoints, align=align, save=save)
        _save_residual_data(errors, contours=contours, save=save)
    elif plots == 1:  # Subplots for profiles/residuals
        if fig is None:
            fig = plt.gcf()
        ax_profiles = fig.add_subplot(211)
        show_profiles(errors, contours=contours, npoints=npoints, align=align, axes=ax_profiles)
        ax_residuals = fig.add_subplot(212)
        show_residuals(errors, contours=contours, axes=ax_residuals)
        if save:
            plt.savefig(save + "-err.png")
    elif plots == 2:  # Separate plots for profiles/residuals
        show_profiles(errors, contours=contours, npoints=npoints, align=align)
        if save:
            plt.savefig(save + "-err1.png")
        plt.figure()
        show_residuals(errors, contours=contours)
        if save:
            plt.savefig(save + "-err2.png")
    else:  # Multiple plots
        profiles, slabs, Q, residuals = errors
        fignum = 1
        for m in profiles.keys():
            plt.figure()
            show_profiles(
                errors=({m: profiles[m]}, {m: slabs[m]}, None, None), contours=contours, npoints=npoints, align=align
            )
            if save:
                plt.savefig(save + "-err%d.png" % fignum)
            fignum += 1
        for m in residuals.keys():
            plt.figure()
            show_residuals(errors=(None, None, {m: Q[m]}, {m: residuals[m]}), contours=contours)
            if save:
                plt.savefig(save + "-err%d.png" % fignum)
            fignum += 1


def show_profiles(errors, align, contours, npoints, axes=None):
    profiles, slabs, _, _ = errors
    if align is not None:
        profiles = align_profiles(profiles, slabs, align)
        _profiles_draw_align_lines(profiles, slabs, align, axes)

    if contours:
        _profiles_contour(profiles, contours, npoints, axes=axes)
    else:
        _profiles_overplot(profiles, axes=axes)


def show_residuals(errors, contours, axes=None):
    _, _, Q, residuals = errors

    if False and contours:
        _residuals_contour(Q, residuals, contours=contours)
    else:
        _residuals_overplot(Q, residuals, axes=axes)


def _save_profile_data(errors, align, contours, npoints, save):
    profiles, slabs, _, _ = errors
    if align is not None:
        profiles = align_profiles(profiles, slabs, align)
    k = 1
    for title, _, group in sorted(
        ((m.name, m.index, group) for m, group in profiles.items()), key=lambda x: (x[0], x[1])
    ):
        # Find limits of all profiles
        z = np.hstack([line[0] for line in group])
        zp = np.linspace(np.min(z), np.max(z), npoints)

        absorbing = any((L[2] != 1e-4).any() for L in group)
        magnetic = len(group[0]) > 3
        twist = magnetic and any((L[4] != BASE_GUIDE_ANGLE).any() for L in group)

        data, columns = _build_profile_matrix(group, 1, zp, contours)
        _write_file(save + "_rho_contour%d.dat" % k, data, title, columns)
        if absorbing:
            data, columns = _build_profile_matrix(group, 2, zp, contours)
            _write_file(save + "_irho_contour%d.dat" % k, data, title, columns)
        if magnetic:
            data, columns = _build_profile_matrix(group, 3, zp, contours)
            _write_file(save + "_rhoM_contour%d.dat" % k, data, title, columns)
        if twist:
            data, columns = _build_profile_matrix(group, 4, zp, contours)
            _write_file(save + "_thetaM_contour%d.dat" % k, data, title, columns)
        k += 1


def _build_profile_matrix(group, index, zp, contours):
    # Interpolate to common z
    fp = np.vstack([np.interp(zp, L[0], L[index]) for L in group])
    # Find quantiles
    q, qval = form_quantiles(fp, contours)
    # Build and return data columns
    columns = ["z", "best"] + list("%g%%" % v for v in 100 * q.flatten())
    data = np.vstack((zp, fp[0], np.reshape(qval, (-1, qval.shape[2]))))
    return data, columns


def _save_residual_data(errors, contours, save):
    _, _, Q, residuals = errors
    k = 1
    for title, _, x, r in sorted(
        [(m.name, m.index, Q[m], v) for m, v in residuals.items()], key=lambda x: (x[0], x[1])
    ):
        q, qval = form_quantiles(r.T, contours)
        # TODO: should have columns for R, dR as well.
        data = np.vstack((x, r[:, 0], np.reshape(qval, (-1, qval.shape[2]))))
        columns = ["q", "best"] + list("%g%%" % v for v in 100 * q.flatten())
        _write_file(save + "_resid_contour%d.dat" % k, data, title, columns)
        k += 1


def _write_file(path, data, title, columns):
    with open(path, "wb") as fid:
        fid.write(asbytes("# " + title + "\n"))
        fid.write(asbytes("# " + " ".join(columns) + "\n"))
        np.savetxt(fid, data.T)


# ===== Plotting functions =====


def dark(color):
    return dhsv(color, dv=-0.2)


def _profiles_overplot(profiles, axes=None):
    for model, group in profiles.items():
        name = model.name
        absorbing = any((L[2] != 1e-4).any() for L in group)
        magnetic = len(group[0]) > 3
        # Note: Use 3 colours per dataset for consistency
        _draw_overplot(group, 1, name + " rho", axes=axes)
        if absorbing:
            _draw_overplot(group, 2, name + " irho", axes=axes)
        else:
            next_color(axes=axes)
        if magnetic:
            _draw_overplot(group, 3, name + " rhoM", axes=axes)
        else:
            next_color(axes=axes)
    _profile_labels(axes=axes)


def _draw_overplot(group, index, label, axes=None):
    import matplotlib.pyplot as plt

    if axes is None:
        axes = plt.gca()
    alpha = 0.1
    color = next_color(axes=axes)
    for L in group[1:]:
        axes.plot(L[0], L[index], "-", color=color, alpha=alpha)
    # Plot best
    L = group[0]
    axes.plot(L[0], L[index], "-", label=label, color=dark(color))


def _profiles_contour(profiles, contours=CONTOURS, npoints=200, axes=None):
    for model, group in profiles.items():
        name = model.name if model.name is not None else "model"
        absorbing = any((L[2] > 1e-4).any() for L in group)
        magnetic = len(group[0]) > 3
        # Find limits of all profiles
        z = np.hstack([line[0] for line in group])
        zp = np.linspace(np.min(z), np.max(z), npoints)
        # Note: Use 3 colours per dataset for consistency
        _draw_contours(group, 1, name + " rho", zp, contours, axes=axes)
        if absorbing:
            _draw_contours(group, 2, name + " irho", zp, contours, axes=axes)
        else:
            next_color(axes=axes)
        if magnetic:
            _draw_contours(group, 3, name + " rhoM", zp, contours, axes=axes)
        else:
            next_color(axes=axes)
    _profile_labels(axes=axes)


def _draw_contours(group, index, label, zp, contours, axes=None):
    import matplotlib.pyplot as plt

    if axes is None:
        axes = plt.gca()
    color = next_color(axes=axes)
    # Interpolate on common z
    fp = np.vstack([np.interp(zp, L[0], L[index]) for L in group])
    # Plot the quantiles
    plot_quantiles(zp, fp, contours, color, axes=axes)
    # Plot the best
    axes.plot(zp, fp[0], "-", label=label, color=dark(color))


def _profile_labels(axes=None):
    import matplotlib.pyplot as plt

    if axes is None:
        axes = plt.gca()
    axes.legend()
    axes.set_xlabel("z (Å)")
    axes.set_ylabel("SLD (10⁻⁶/Å²)")


def _residuals_overplot(Q, residuals, axes=None):
    import matplotlib.pyplot as plt

    if axes is None:
        axes = plt.gca()
    alpha = 0.4
    shift = 0
    for m, r in residuals.items():
        color = next_color(axes=axes)
        axes.plot(Q[m], shift + r[:, 1:], ".", markersize=1, color=color, alpha=alpha)
        axes.plot(Q[m], shift + r[:, 0], ".", label=m.name, markersize=1, color=dark(color))
        # Use 3 colours from cycle so reflectivity matches rho for each dataset
        next_color(axes=axes)
        next_color(axes=axes)
        shift += 5
    _residuals_labels(axes=axes)


def _residuals_contour(Q, residuals, contours=CONTOURS):
    import matplotlib.pyplot as plt

    shift = 0
    for m, r in residuals.items():
        color = next_color()
        plot_quantiles(Q[m], shift + r.T, contours, color)
        plt.plot(Q[m], shift + r[:, 0], ".", label=m.name, markersize=1, color=dark(color))
        # Use 3 colours from cycle so reflectivity matches rho for each dataset
        next_color()
        next_color()
        shift += 5
    _residuals_labels()


def _residuals_labels(axes=None):
    import matplotlib.pyplot as plt

    if axes is None:
        axes = plt.gca()
    axes.legend()
    axes.set_xlabel("Q (1/Å)")
    axes.set_ylabel("Residuals")


def _profiles_draw_align_lines(profiles, slabs, align, axes):
    for i, m in enumerate(profiles.keys()):
        t1_offset = _find_offset(slabs[m][0], align) if align != "auto" else None
        if t1_offset is not None:
            axes.axvline(x=t1_offset, color="grey", label=f"{m}:{i}")


# ==== Helper functions =====


def _align_profile_set(profiles, slabs, align):
    """
    Align all profiles to the first profile.
    """
    p1 = profiles[0]
    t1_offset = _find_offset(slabs[0], align) if align != "auto" else None
    offsets = [0]
    for p2, t2 in zip(profiles[1:], slabs[1:]):
        offsets.append(_align_profile_pair(p1[0], p1[1], t1_offset, p2[0], p2[1], t2, align))
    profiles = [tuple([group[0] + offset] + list(group[1:])) for offset, group in zip(offsets, profiles)]
    return profiles


def _align_profile_pair(z1, r1, t1_offset, z2, r2, t2, align):
    """
    Use crosscorrelation to align r1 and r2.
    """
    if align == "auto":
        import scipy.signal

        # Assume z1, z2 have the same step size
        n2 = len(r2)
        idx = np.argmax(scipy.signal.correlate(r1, r2, "full"))
        if idx < n2:
            offset = z2[(n2 - 1) - idx] - z1[0]
        else:
            offset = z2[0] - z1[idx - (n2 - 1)]
        return -offset
    else:
        return -(_find_offset(t2, align) - t1_offset)


def _find_offset(v, align):
    """
    Find the offset of k.p, where k is the interface number and p is the
    distance into that interface.

    This may even work for interfaces defined from the left, such as
    -1.5 to specify the middle of the final layer.
    """
    idx = int(align)
    offset = np.sum(v[:idx]) + np.sum((align - idx) * v[idx : idx + 1])
    # print offset, idx, v[:idx], align
    return offset



import argparse

import numpy as np


def get_optimal_single_oversampling(model, tolerance=0.05, max_oversampling=201, seed=1, verbose=False):
    """
    Determine how much oversampling is required to adequately calculate resolution smearing

    Algorithm:
      - the model is oversampled to a very high number (max_oversampling)
      - smeared R is calculated at max_oversampling and stored as reference "ideal" R
      - oversampling is iterated (starting at 1), and smeared R is calculated and compared to "ideal" R
      - points are "within tolerance" if (R(Q) - R_ref(Q)) / dR(Q) < tolerance
      - the value of oversampling at which a point becomes within tolerance is recorded for that point
      - when all points are within tolerance the loop ends and the function reports the recommended oversampling

    Args:
        model (refl1d.experiment.Experiment): an Experiment, containing probe and sample
        tolerance (float, optional): allowed deviation of R from ideal R (multiplied by dR). Defaults to 0.05
        max_oversampling (int, optional): A very high oversampling that is expected to exceed
            the requirements for support (used to generate reference R). Defaults to 201.

    Returns:
        oversampling (int): suggested oversampling to get all R within tolerance
        optimal_oversampling: for each probe, per-Q array of oversampling needed to get R(Q) within tolerance
    """
    # get a list of probes, which will have length one for unpolarized:
    probes = model.probe.xs if hasattr(model.probe, "xs") else [model.probe]
    # sample = model.sample

    # initialize the per-Q recommended oversampling to max_oversampling
    optimal_oversampling = [None if p is None else np.ones_like(p.dR, dtype=int) * max_oversampling for p in probes]
    Q = []
    model.probe.oversample(max_oversampling, seed=seed)
    model._cache = {}
    R_ref = model.reflectivity()
    if not isinstance(R_ref, list):
        R_ref = [R_ref]

    max_diff = np.inf
    oversampling = 0
    while max_diff > tolerance and oversampling < max_oversampling:
        oversampling += 1
        if verbose:
            print("trying oversampling = {:d}".format(oversampling), end="\r")
        model.probe.oversample(oversampling)
        model._cache = {}
        R = model.reflectivity()
        if not isinstance(R, list):
            R = [R]
        current_max_diff = 0
        for oos, p, r, r_ref in zip(optimal_oversampling, probes, R, R_ref):
            if p is None:
                Q.append(None)
                continue

            Q.append(p.Q)
            diff = np.abs(r[1] - r_ref[1]) / p.dR
            current_max_diff = max(current_max_diff, diff.max())

            # if an R value is now within tolerance, and the optimal_oversampling has not been set
            # (which is indicated by the fact that it is still set to max_oversampling), then
            # set it to the current oversampling:
            to_set = np.logical_and(diff < tolerance, oos == max_oversampling)
            oos[to_set] = oversampling

            # if an R value is now out of tolerance, make sure that the optimal_oversampling is reset
            # to the "unset" value of max_oversampling
            to_reset = diff > tolerance
            oos[to_reset] = max_oversampling

        max_diff = current_max_diff  # to be checked against tolerance on the next iteration

    # reset model cache again before leaving
    model._cache = {}
    if verbose:
        print("Recommended oversampling: {:d}".format(oversampling))

    return oversampling, optimal_oversampling, Q


def analyze_fitproblem(problem, tolerance=0.05, max_oversampling=201, seed=1, plot=False):
    if plot:
        from matplotlib import pyplot as plt
    models = problem.models if hasattr(problem, "models") else [problem]
    oversampling = []
    local_oversampling = []
    for i_model, model in enumerate(models):
        print("model: {:d}".format(i_model))
        oversampling_i = []
        local_oversampling_i = []
        oversampling.append(oversampling_i)
        local_oversampling.append(local_oversampling_i)
        parts = model.parts if hasattr(model, "parts") else [model]
        for i_part, part in enumerate(parts):
            # print(part)
            oversampling_ii, local_oversampling_ii, Q = get_optimal_single_oversampling(
                part, tolerance, max_oversampling, seed=seed
            )
            print("\tpart: {:d}, oversampling: {:d}".format(i_part, oversampling_ii))
            oversampling_i.append(oversampling_ii)
            local_oversampling_i.append(local_oversampling_ii)
            if plot:
                plt.figure()
                for i_oos, (oos, qq) in enumerate(zip(local_oversampling_ii, Q)):
                    if oos is not None:
                        plt.plot(qq, oos, label="xs: {:d}".format(i_oos))
                plt.title("model: {:d}, part: {:d}".format(i_model, i_part))
                plt.xlabel("Q (inv. A)")
                plt.ylabel("required oversampling")
                plt.legend()

        # there is one probe instance shared between parts in MixedExperiment: use
        # largest recommended oversampling.
        parts[0].probe.oversample(max(oversampling_i))

    if plot:
        plt.show()

    return oversampling, local_oversampling


def main():
    import sys
    from bumps.cli import load_model, load_best

    parser = argparse.ArgumentParser(
        description="""
        Algorithm:
      - the model is oversampled to a very high number (max_oversampling)
      - smeared R is calculated at max_oversampling and stored as reference "ideal" R
      - oversampling is iterated (starting at 1), and smeared R is calculated and compared to "ideal" R
      - points are "within tolerance" if (R(Q) - R_ref(Q)) / dR(Q) < tolerance
      - the value of oversampling at which a point becomes within tolerance is recorded for that point
      - when all points are within tolerance the loop ends and the function reports the recommended oversampling
      """,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--plot", action="store_true", help="plot optimal oversampling for each probe (default = False)"
    )
    parser.add_argument(
        "-t",
        "--tolerance",
        type=float,
        default=0.05,
        help="Tolerance for expression: (R - R_ideal)/dR < tolerance (default=0.05)",
    )
    parser.add_argument(
        "-m",
        "--max_oversampling",
        type=int,
        default=201,
        help="Max oversampling (also used to calculate R_ideal; default=201)",
    )
    parser.add_argument("-p", "--pars", type=str, default="", help="retrieve starting point from .par file")

    parser.add_argument("modelfile", type=str, nargs=1, help="refl1d model file")
    parser.add_argument("modelopts", type=str, nargs="*", help="options passed to the model")

    opts = parser.parse_args(None if sys.argv[1:] else ["-h"])

    problem = load_model(opts.modelfile[0], model_options=opts.modelopts)
    if opts.pars:
        load_best(problem, opts.pars)

    analyze_fitproblem(problem, opts.tolerance, opts.max_oversampling, opts.plot)


if __name__ == "__main__":
    main()



from copy import deepcopy
from typing import Optional, Tuple
import re

from bumps.serialize import TYPE_KEY
from .. import __schema_version__ as CURRENT_SCHEMA_VERSION


def migrate(
    serialized: dict, from_version: Optional[str] = None, to_version: Optional[str] = CURRENT_SCHEMA_VERSION
) -> Tuple[str, dict]:
    """
    Migrate a serialized object from one version to another
    By default, the `from_version` is determined by inspection of the serialized object.
    This is overriden by setting the `from_version` keyword argument

    Also by default, the target version is the current schema, which can be overriden with
    the `to_version` keyword argument
    """

    if from_version is None:
        libraries = serialized.get("libraries", {})
        from_version = libraries.get("refl1d", {}).get("schema_version", "0")

    current_version = from_version
    while current_version != to_version:
        print(f"migrating refl1d from schema: {current_version}")
        current_version, serialized = MIGRATIONS[current_version](serialized)

    return current_version, serialized


def _migrate_0_to_1(serialized: dict):
    MAPPINGS = {
        "refl1d.abeles": "refl1d.probe.abeles",
        "refl1d.fresnel": "refl1d.probe.fresnel",
        "refl1d.instrument": "refl1d.probe.instrument",
        "refl1d.oversampling": "refl1d.probe.oversampling",
        "refl1d.probe": "refl1d.probe.probe",
        "refl1d.cheby": "refl1d.sample.cheby",
        "refl1d.model": "refl1d.sample.layers",
        "refl1d.material": "refl1d.sample.material",
        "refl1d.materialdb": "refl1d.sample.materialdb",
        "refl1d.magnetic": "refl1d.sample.magnetic",
        "refl1d.magnetism": "refl1d.sample.magnetism",
        "refl1d.reflectivity": "refl1d.sample.reflectivity",
        "refl1d.polymer": "refl1d.sample.polymer",
        "refl1d.mono": "refl1d.sample.mono",
        "refl1d.flayer": "refl1d.sample.flayer",
        "refl1d.util": "refl1d.utils.util",
        "refl1d.support": "refl1d.utils.support",
    }

    def remap(obj):
        if isinstance(obj, dict):
            if TYPE_KEY in obj:
                classname: str = obj[TYPE_KEY]
                first_dot_pair = ".".join(classname.split(".")[:2])
                if first_dot_pair in MAPPINGS:
                    obj[TYPE_KEY] = classname.replace(first_dot_pair, MAPPINGS[first_dot_pair], 1)
            for v in obj.values():
                remap(v)
        elif isinstance(obj, list):
            for v in obj:
                remap(v)

    output = deepcopy(serialized)
    remap(output)

    return "1", output


MIGRATIONS = {
    "0": _migrate_0_to_1,
}



"""
Testing experiment serialization
"""

import os
import unittest

import numpy as np

from refl1d.names import QProbe, Slab, SLD, Parameter, Experiment, NeutronProbe, PolarizedNeutronProbe, Magnetism


class ExperimentJsonTest(unittest.TestCase):
    """Simple Experiment serialization test"""

    def test_save_json(self):
        """Check that we can produce a json file"""
        q_values = np.logspace(-2.1, -0.7, 100)
        # Fake data array
        zeros = np.zeros(len(q_values))
        ones = np.ones(len(q_values))
        # Q-resolution array
        q_resolution = q_values * 0.02 + 0.0001

        probe = QProbe(q_values, q_resolution, data=(zeros, ones))

        # Note: I don't use the usual shorthand to define the sample on purpose,
        #       so I see explicitly what I'm using.
        sample = (
            Slab(material=SLD(name="Si", rho=2.07, irho=0.0))
            | Slab(material=SLD(name="Cu", rho=6.5, irho=0.0), thickness=130, interface=15)
            | Slab(material=SLD(name="air", rho=0, irho=0.0))
        )

        sample["Cu"].thickness.range(90.0, 200.0)

        probe.intensity = Parameter(value=1.0, name="normalization")
        probe.background = Parameter(value=0.0, name="background")

        expt = Experiment(probe=probe, sample=sample)
        expt.save("output")

        self.assertTrue(os.path.isfile("output-expt.json"))


class ExperimentMagneticSimulateTest(unittest.TestCase):
    """Test the simulation functionality"""

    def setUp(self):
        self.q_values = np.logspace(-2.1, -0.6, 100)
        L = 4.75
        dL = 0.0475
        dT = 0.01

        xs = [NeutronProbe(T=self.q_values, dT=dT, L=L, dL=dL) for _ in range(4)]
        probe = PolarizedNeutronProbe(xs)

        # Note: I don't use the usual shorthand to define the sample on purpose,
        #       so I see explicitly what I'm using.
        # Note: 'MagneticSlab' is deprecated.  Using magnetism=... instead.
        sample = (
            Slab(material=SLD(name="Si", rho=2.07, irho=0.0))
            | Slab(material=SLD(name="Cu", rho=6.5, irho=0.0), thickness=130, interface=15)
            | Slab(
                SLD(name="Stuff", rho=4.0, irho=0.0),
                thickness=50.0,
                interface=1.0,
                magnetism=Magnetism(rhoM=0.2, thetaM=270),
            )
            | Slab(material=SLD(name="air", rho=0, irho=0.0))
        )

        sample["Cu"].thickness.range(90.0, 200.0)

        probe.intensity = Parameter(value=1.0, name="normalization")
        probe.background = Parameter(value=0.0, name="background")

        self.expt = Experiment(probe=probe, sample=sample)

        # Reference
        self.expt.simulate_data(noise=1e-11)
        self.r_i = self.expt.probe.pp.R[0]
        self.r_f = self.expt.probe.pp.R[-1]

    def test_none_noise_with_mag(self):
        """Provide None for noise with a magnetic sample"""

        self.expt.simulate_data(noise=None)
        self.assertEqual(len(self.expt.probe.pp.dR), len(self.q_values))

        ratio_i = self.expt.probe.pp.dR[0] / self.r_i
        ratio_f = self.expt.probe.pp.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.0)
        self.assertAlmostEqual(ratio_f, 0.0)

    def test_four_none_noise_with_mag(self):
        """Provide [None]*4 for noise with a magnetic sample"""

        self.expt.simulate_data(noise=[None] * 4)
        self.assertEqual(len(self.expt.probe.pp.dR), len(self.q_values))

        ratio_i = self.expt.probe.pp.dR[0] / self.r_i
        ratio_f = self.expt.probe.pp.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.0)
        self.assertAlmostEqual(ratio_f, 0.0)

    def test_noise_scalar_with_mag(self):
        """Provide a scalar of dR with a magnetic sample"""

        self.expt.simulate_data(noise=2.5)
        self.assertEqual(len(self.expt.probe.pp.dR), len(self.q_values))

        ratio_i = self.expt.probe.pp.dR[0] / self.r_i
        ratio_f = self.expt.probe.pp.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.025)
        self.assertAlmostEqual(ratio_f, 0.025)

    def test_four_noise_scalar_with_mag(self):
        """Provide four scalars of dR with a magnetic sample"""

        self.expt.simulate_data(noise=[2.5, 2.5, 2.5, 2.5])
        self.assertEqual(len(self.expt.probe.pp.dR), len(self.q_values))

        ratio_i = self.expt.probe.pp.dR[0] / self.r_i
        ratio_f = self.expt.probe.pp.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.025)
        self.assertAlmostEqual(ratio_f, 0.025)

    # unittest.skip("ambiguous interface not supported")
    def not_test_noise_array_with_mag(self):
        """Provide a dR array with a magnetic sample"""
        m = 10.0 / (self.q_values[-1] - self.q_values[0])
        b = 1 - m * self.q_values[0]
        _noise = m * self.q_values + b

        self.expt.simulate_data(noise=_noise)
        self.assertEqual(len(self.expt.probe.pp.dR), len(self.q_values))

        ratio_i = self.expt.probe.pp.dR[0] / self.r_i
        ratio_f = self.expt.probe.pp.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.01)
        self.assertAlmostEqual(ratio_f, 0.11)

    def test_four_noise_array_with_mag(self):
        """Provide four dR arrays with a magnetic sample"""
        m = 10.0 / (self.q_values[-1] - self.q_values[0])
        b = 1 - m * self.q_values[0]
        _noise = m * self.q_values + b

        self.expt.simulate_data(noise=[_noise, _noise, _noise, _noise])
        self.assertEqual(len(self.expt.probe.pp.dR), len(self.q_values))

        ratio_i = self.expt.probe.pp.dR[0] / self.r_i
        ratio_f = self.expt.probe.pp.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.01)
        self.assertAlmostEqual(ratio_f, 0.11)


class ExperimentNonMagSimulateTest(unittest.TestCase):
    """Test the simulation functionality"""

    def setUp(self):
        self.q_values = np.logspace(-2.1, -0.6, 100)
        L = 4.75
        dL = 0.0475
        dT = 0.01

        probe = NeutronProbe(T=self.q_values, dT=dT, L=L, dL=dL)

        # Note: I don't use the usual shorthand to define the sample on purpose,
        #       so I see explicitly what I'm using.
        sample = (
            Slab(material=SLD(name="Si", rho=2.07, irho=0.0))
            | Slab(material=SLD(name="Cu", rho=6.5, irho=0.0), thickness=130, interface=15)
            | Slab(material=SLD(name="air", rho=0, irho=0.0))
        )

        sample["Cu"].thickness.range(90.0, 200.0)

        probe.intensity = Parameter(value=1.0, name="normalization")
        probe.background = Parameter(value=0.0, name="background")

        self.expt = Experiment(probe=probe, sample=sample)

        # Reference
        self.expt.simulate_data(noise=1e-11)
        self.r_i = self.expt.probe.R[0]
        self.r_f = self.expt.probe.R[-1]

    def test_noise_scalar_non_mag(self):
        """Provide a scalar of dR with a non-magnetic sample"""

        self.expt.simulate_data(noise=2.5)
        self.assertEqual(len(self.expt.probe.dR), len(self.q_values))

        ratio_i = self.expt.probe.dR[0] / self.r_i
        ratio_f = self.expt.probe.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.025)
        self.assertAlmostEqual(ratio_f, 0.025)

    def test_noise_array_non_mag(self):
        """Provide an array of dR with a non-magnetic sample"""
        m = 10.0 / (self.q_values[-1] - self.q_values[0])
        b = 1 - m * self.q_values[0]
        _noise = m * self.q_values + b
        # print("noise input", _noise)

        self.expt.simulate_data(noise=_noise)
        self.assertEqual(len(self.expt.probe.dR), len(self.q_values))

        ratio_i = self.expt.probe.dR[0] / self.r_i
        ratio_f = self.expt.probe.dR[-1] / self.r_f

        self.assertAlmostEqual(ratio_i, 0.01)
        self.assertAlmostEqual(ratio_f, 0.11)


if __name__ == "__main__":
    unittest.main()



# import asyncio
# from aiohttp import web
from dataclasses import dataclass

# import json
from pathlib import Path
# import socket
# from typing import Union, Dict, List, Optional

# import numpy as np
from bumps.webview.server import webserver
from bumps.webview.server.webserver import (
    # start_app,
    # sio,
    main,
    # create_server_task,
    # display_inline_jupyter,
    # open_tab_link,
)

# Register the refl1d model loader
from refl1d.bumps_interface import fitplugin
import bumps.cli

bumps.cli.install_plugin(fitplugin)

from . import api  # use side-effects to register refl1d functions
# from .profile_plot import plot_sld_profile_plotly

webserver.CLIENT_PATH = Path(__file__).parent.parent / "client"


@dataclass
class Refl1DOptions(webserver.BumpsOptions):
    serializer: webserver.SERIALIZERS = "dataclass"
    headless: bool = True


webserver.OPTIONS_CLASS = Refl1DOptions
webserver.APPLICATION_NAME = "refl1d"

if __name__ == "__main__":
    main()



"""
Optical matrix form of the reflectivity calculation.

O.S. Heavens, Optical Properties of Thin Solid Films

This is a pure python implementation of reflectometry provided for
convenience when a compiler is not available.  The refl1d
application uses reflmodule to compute reflectivity.
"""

from numpy import asarray, isscalar, empty, ndarray, ones, ones_like
from numpy import sqrt, exp, pi


def refl(kz, depth, rho, irho=0, sigma=0, rho_index=None):
    r"""
    Reflectometry as a function of kz for a set of slabs.

    :Parameters:

    *kz* : float[n] | |1/Ang|
        Scattering vector $2\pi\sin(\theta)/\lambda$. This is $\tfrac12 Q_z$.
    *depth* :  float[m] | |Ang|
        thickness of each layer.  The thickness of the incident medium
        and substrate are ignored.
    *rho*, *irho* :  float[n, k] | |1e-6/Ang^2|
        real and imaginary scattering length density for each layer for each kz
        Note: absorption cross section mu = 2 irho/lambda
    *sigma* : float[m-1] | |Ang|
        interfacial roughness.  This is the roughness between a layer
        and the subsequent layer.  There is no interface associated
        with the substrate.  The sigma array should have at least m-1
        entries, though it may have m with the last entry ignored.
    *rho_index* : int[m]
        index into rho vector for each kz

    Slabs are ordered with the surface SLD at index 0 and substrate at
    index -1, or reversed if kz < 0.
    """
    if isscalar(kz):
        kz = asarray([kz], "d")

    m = len(depth)

    # Make everything into arrays
    depth = asarray(depth, "d")
    rho = asarray(rho, "d")
    irho = irho * ones_like(rho) if isscalar(irho) else asarray(irho, "d")
    sigma = sigma * ones(m - 1, "d") if isscalar(sigma) else asarray(sigma, "d")

    # Repeat rho, irho columns as needed
    if rho_index is not None:
        rho = rho[rho_index, :]
        irho = irho[rho_index, :]
    elif len(rho.shape) == 1:
        rho = rho[None, :]
        irho = irho[None, :]

    # Force the correct branch cut for sqrt below the critical edge
    irho = abs(irho) + 1e-30

    ## For kz < 0 we need to reverse the order of the layers
    ## Note that the interface array sigma is conceptually one
    ## shorter than rho, mu so when reversing it, start at n-1.
    ## This allows the caller to provide an array of length n
    ## corresponding to rho, mu or of length n-1.
    r = empty(len(kz), "D")
    r[kz >= 1e-10] = _calc(kz[kz >= 1e-10], depth, rho, irho, sigma)
    r[kz <= 1e-10] = _calc(-kz[kz <= 1e-10], depth[::-1], rho[:, ::-1], irho[:, ::-1], sigma[m - 2 :: -1])
    r[abs(kz) < 1e-10] = -1
    return r


def _calc(kz: ndarray, depth: ndarray, rho: ndarray, irho: ndarray, sigma: ndarray) -> ndarray:
    """Reflectivity as a function of kz for a set of slabs."""
    if len(kz) == 0:
        return kz

    # Complex index of refraction is relative to the incident medium.
    # We can get the same effect using kz_rel^2 = kz^2 + 4*pi*rho_o
    # in place of kz^2, and ignoring rho_o
    kz_sq = kz**2 + 4e-6 * pi * rho[:, 0]
    k = kz

    # According to Heavens, the initial matrix should be [ 1 F; F 1],
    # which we do by setting B=I and M0 to [1 F; F 1].  An extra matrix
    # multiply versus some coding convenience.
    B11 = 1
    B22 = 1
    B21 = 0
    B12 = 0
    for i in range(0, len(depth) - 1):
        k_next = sqrt(kz_sq - 4e-6 * pi * (rho[:, i + 1] + 1j * irho[:, i + 1]))
        F = (k - k_next) / (k + k_next)
        F *= exp(-2 * k * k_next * sigma[i] ** 2)
        # print("==== layer", i)
        # print("kz:", kz.real)
        # print("k:", k.real)
        # print("k_next:", k_next.real)
        # print("F:", F.real)
        # print("rho:", rho[:, i+1])
        # print("irho:", irho[:, i+1])
        # print("d:", depth[i], "sigma:", sigma[i])
        M11 = exp(1j * k * depth[i]) if i > 0 else 1
        M22 = exp(-1j * k * depth[i]) if i > 0 else 1
        M21 = F * M11
        M12 = F * M22
        C1 = B11 * M11 + B21 * M12
        C2 = B11 * M21 + B21 * M22
        B11 = C1
        B21 = C2
        C1 = B12 * M11 + B22 * M12
        C2 = B12 * M21 + B22 * M22
        B12 = C1
        B22 = C2
        k = k_next
        # print("B11:", B11)
        # print("B22:", B22)
        # print("B21:", B21)
        # print("B12:", B12)
        # print("1-det:", 1 - (B11*B22 - B21*B12))

    r = B12 / B11
    return r


def test():
    import numpy as np

    np.set_printoptions(linewidth=10000)

    q = np.linspace(-0.3, 0.3, 6)
    # q = np.linspace(0.1, 0.3, 3)
    layers = [
        # depth rho irho sigma
        [0, 1.0, 0.0, 10.0],
        [200, 2.0, 1.0, 10.0],
        [200, 4.0, 0.0, 10.0],
        [0, 2.0, 0.0, 0.0],
    ]
    # add absorption
    # layers[1][2] = 1.0

    depth, rho, irho, sigma = zip(*layers)
    r = refl(q / 2, depth, rho, irho=irho, sigma=sigma)
    print("q", q)
    print("r", r)
    # print("r^2", abs(r**2))


if __name__ == "__main__":
    test()



"""
Inhomogeneous samples

In the presence of samples with short range order on scale of the coherence
length of the probe in the plane, but long range disorder following some
distribution of parameter values, the reflectivity can be computed from
a weighted incoherent sum of the reflectivities for different values of
the parameter.

DistristributionExperiment allows the model to be computed for a single
varying parameter.  Multi-parameter dispersion models are not available.
"""

from bumps.parameter import Parameter
import numpy as np

from .experiment import ExperimentBase


class Weights(object):
    """
    Parameterized distribution for use in DistributionExperiment.

    To support non-uniform experiments, we must bin the possible values
    for the parameter and compute the theory function for one parameter
    value per bin.  The weighted sum of the resulting theory functions
    is the value that we compare to the data.

    Performing this analysis requires a cumulative density function which
    can return the integrated value of the probability density from -inf
    to x.  The total density in each bin is then the difference between
    the cumulative densities at the edges.  If the distribution is wider
    than the range, then the tails need to be truncated and the bins
    reweighted to a total density of 1, or the tail density can be added
    to the first and last bins.  Weights of zero are not returned.  Note
    that if the tails are truncated, this may result in no weights being
    returned.

    The vector *edges* contains the bin edges for the distribution.  The
    function *cdf* returns the cumulative density function at the edges.
    The *cdf* function must implement the scipy.stats interface, with
    function signature f(x, a1, a2, ..., loc=0, scale=1).  The list *args*
    defines the arguments a1, a2, etc.  The underlying parameters are
    available as args[i].  Similarly, *loc* and *scale* define the
    distribution center and width.  Use *truncated=False* if you want
    the distribution tails to be included in the weights.

    SciPy distribution D is used by specifying cdf=scipy.stats.D.cdf.
    Useful distributions include::

        norm      Gaussian distribution.
        halfnorm  Right half of a gaussian.
        triang    Triangle distribution from loc up to loc+args[0]*scale
                  and down to loc+scale.  Use loc=edges[0], scale=edges[-1]
                  and args=[0.5] to define a symmetric triangle in the range
                  of parameter P.
        uniform   Flat from loc to loc+scale. Use loc=edges[0], scale=edges[-1]
                  to define P as uniform over the range.

    """

    def __init__(self, edges=None, cdf=None, args=(), loc=None, scale=None, truncated=True):
        self.edges = np.asarray(edges)
        self.cdf = cdf
        self.truncated = truncated
        self.loc = Parameter.default(loc)
        self.scale = Parameter.default(scale)
        self.args = [Parameter.default(a) for a in args]

    def parameters(self):
        return {"args": self.args, "loc": self.loc, "scale": self.scale}

    def __iter__(self):
        # Find weights and normalize the sum to 1
        centers = (self.edges[:-1] + self.edges[1:]) / 2
        loc = self.loc.value
        scale = self.scale.value
        args = [p.value for p in self.args]
        cumulative_weights = self.cdf(self.edges, *args, loc=loc, scale=scale)
        if not self.truncated:
            cumulative_weights[0], cumulative_weights[-1] = 0, 1
        relative_weights = cumulative_weights[1:] - cumulative_weights[:-1]
        total_weight = np.sum(relative_weights)
        if total_weight == 0:
            return iter([])
        else:
            weights = relative_weights / total_weight
            idx = weights > 0
            return iter(zip(centers[idx], weights[idx]))


class DistributionExperiment(ExperimentBase):
    """
    Compute reflectivity from a non-uniform sample.

    *P* is the target parameter for the model, which takes on the values
    from *distribution* in the context of the *experiment*.  The result
    is the weighted sum of the theory curves after setting *P.value* to
    each distribution value. Clearly, *P* should not be a fitted parameter,
    but the remaining experiment parameters can be fitted, as can the
    parameters of the distribution.

    If *coherent* is true, then the reflectivity of the mixture is computed
    from the coherent sum rather than the incoherent sum.

    See :class:`Weights` for a description of how to set up the distribution.
    """

    def __init__(self, experiment=None, P=None, distribution=None, coherent=False):
        self.P = P
        self.distribution = distribution
        self.experiment = experiment
        self.probe = self.experiment.probe
        self.coherent = coherent
        self._substrate = self.experiment.sample[0].material
        self._surface = self.experiment.sample[-1].material
        self._cache = {}  # Cache calculated profiles/reflectivities
        self._name = None

    def parameters(self):
        return {
            "distribution": self.distribution.parameters(),
            "experiment": self.experiment.parameters(),
        }

    def reflectivity(self, resolution=True, interpolation=0):
        key = ("reflectivity", resolution, interpolation)
        if key not in self._cache:
            calc_R = 0
            for x, w in self.distribution:
                if w > 0:
                    self.P.value = x
                    self.experiment.update()
                    Qx, Rx = self.experiment._reflamp()
                    if self.coherent:
                        calc_R += w * Rx
                    else:
                        calc_R += w * abs(Rx) ** 2
            if self.coherent:
                calc_R = abs(calc_R) ** 2
            Q, R = self.probe.apply_beam(Qx, calc_R, resolution=resolution, interpolation=interpolation)
            self._cache[key] = Q, R
        return self._cache[key]

    def _max_P(self):
        x, w = zip(*self.distribution)
        idx = np.argmax(w)
        return x[idx]

    def smooth_profile(self, dz=1):
        """
        Compute a density profile for the material
        """
        key = "smooth_profile", dz
        if key not in self._cache:
            P = self._max_P()
            if self.P.value != P:
                self.P.value = P
                self.experiment.update()
            self._cache[key] = self.experiment.smooth_profile(dz=dz)
        return self._cache[key]

    def step_profile(self):
        """
        Compute a scattering length density profile
        """
        key = "step_profile"
        if key not in self._cache:
            P = self._max_P()
            if self.P.value != P:
                self.P.value = P
                self.experiment.update()
            self._cache[key] = self.experiment.step_profile()
        return self._cache[key]

    def plot_profile(self, plot_shift=0.0):
        from bumps.plotutil import auto_shift
        import matplotlib.pyplot as plt

        trans = auto_shift(plot_shift)
        z, rho, irho = self.step_profile()
        plt.plot(z, rho, "-g", z, irho, "-b", transform=trans)
        z, rho, irho = self.smooth_profile()
        plt.plot(z, rho, ":g", z, irho, ":b", transform=trans)
        plt.legend(["rho", "irho"])

    def plot_weights(self):
        import matplotlib.pyplot as plt

        x, w = zip(*self.distribution)
        plt.stem(x, 100 * np.array(w))
        plt.title("Weight distribution")
        plt.xlabel(self.P.name)
        plt.ylabel("Percentage")



# This program is in the public domain
# Author: Paul Kienzle
"""
SNS data loaders

The following instruments are defined::

    Liquids, Magnetic

These are :class:`resolution.Pulsed` classes tuned with
default instrument parameters and loaders for reduced SNS data.
See :mod:`resolution` for details.
"""

import math
import re

import numpy as np
from bumps.data import parse_file

from ...probe.probe import make_probe
from .. import resolution
from ..instrument import Pulsed
from .rebin import rebin

## Estimated intensity vs. wavelength for liquids reflectometer
LIQUIDS_FEATHER = np.array(
    [
        (2.02555, 20.6369),
        (2.29927, 23.6943),
        (2.57299, 23.6943),
        (2.87409, 21.1146),
        (3.22993, 15.5732),
        (3.58577, 12.8981),
        (4.07847, 9.4586),
        (4.5438, 6.59236),
        (5.11861, 4.68153),
        (5.7208, 3.05732),
        (6.37774, 1.91083),
        (7.19891, 1.24204),
        (8.04745, 0.955414),
        (9.06022, 0.573248),
        (10.1825, 0.477707),
        (11.4142, 0.382166),
        (12.8102, 0.191083),
        (14.3431, 0.286624),
    ]
).T


def load(filename, instrument=None, **kw):
    """
    Return a probe for SNS data.
    """
    if instrument is None:
        instrument = Pulsed()
    header, data = parse_sns_file(filename)
    header.update(**kw)  # calling parameters override what's in the file.
    # print "\n".join(k+":"+str(v) for k, v in header.items())
    # Guess what kind of data we have
    if has_columns(header, ("Q", "dQ", "R", "dR", "L")):
        probe = QRL_to_data(instrument, header, data)
    elif has_columns(header, ("time_of_flight", "data", "Sigma")):
        probe = TOF_to_data(instrument, header, data)
    else:
        raise IOError("Unknown columns: " + ", ".join(header["columns"]))
    probe.title = header["title"]
    probe.date = header["date"]
    probe.instrument = header["instrument"]
    return probe


def has_columns(header, v):
    return len(header["columns"]) == len(v) and all(ci == si for ci, si in zip(header["columns"], v))


def QRL_to_data(instrument, header, data):
    """
    Convert data to T, L, R
    """
    Q, dQ, R, dR, L = data
    dL = resolution.binwidths(L)
    if "angle" in header and "slits_at_Tlo" in header:
        T = header.pop("angle", header.pop("T", None))
        probe = instrument.probe(L=L, dL=dL, T=T, data=(R, dR), **header)
    else:
        T = resolution.QL2T(Q[0], L[0])
        dT = resolution.dQdL2dT(Q[0], dQ[0], L[0], dL[0])
        probe = make_probe(T=T, dT=dT, L=L, dL=dL, data=(R, dR), **header)
    return probe


def TOF_to_data(instrument, header, data):
    """
    Convert TOF data to neutron probe.

    Wavelength is set from the average of the times at the edges of the
    bins, not the average of the wavelengths.  Wavelength resolution is
    set assuming the wavelength at the edges of the bins defines the
    full width at half maximum.

    The correct answer is to look at the wavelength distribution within
    the bin including effects of pulse width and intensity as a function
    wavelength and use that distribution, or a gaussian approximation
    thereof, when computing the resolution effects.
    """
    TOF, R, dR = data
    Ledge = resolution.TOF2L(instrument.d_moderator, TOF)
    L = resolution.TOF2L(instrument.d_moderator, (TOF[:-1] + TOF[1:]) / 2)
    dL = (Ledge[1:] - Ledge[:-1]) / 2.35  # FWHM is 2.35 sigma
    R = R[:-1]
    dR = dR[:-1]
    min_time, max_time = header.get("TOF_range", instrument.TOF_range)
    keep = np.isfinite(R) & np.isfinite(dR) & (TOF[:-1] >= min_time) & (TOF[1:] <= max_time)
    L, dL, R, dR = [v[keep] for v in (L, dL, R, dR)]
    T = np.array([header.get("angle", header.get("T", None))], "d")
    T, dT, L, dL = instrument.resolution(L=L, dL=dL, T=T, **header)
    probe = make_probe(T=T, dT=dT, L=L, dL=dL, data=(R, dR), **header)
    return probe


def parse_sns_file(filename):
    """
    Parse SNS reduced data, returning *header* and *data*.

    *header* dictionary of fields such as 'data', 'title', 'instrument'
    *data* 2D array of data
    """
    raw_header, data = parse_file(filename)
    header = {}

    # guess instrument from file name
    original_file = raw_header.get("F", "unknown")
    if "REF_L" in original_file:
        instrument = "Liquids"
    elif "REF_M" in original_file:
        instrument = "Magnetic"
    else:
        instrument = "unknown"
    header["instrument"] = instrument
    header["filename"] = original_file
    header["radiation"] = "neutron"

    # Plug in default instrument values for slits
    if "instrument" in header and header["instrument"] in INSTRUMENTS:
        instrument = INSTRUMENTS[header["instrument"]]
        header["d_s1"] = instrument.d_s1
        header["d_s2"] = instrument.d_s2

    # Date-time field for the file
    header["date"] = raw_header.get("D", "")

    # Column names and units
    columnpat = re.compile(r"(?P<name>\w+)\((?P<units>[^)]*)\)")
    columns, units = zip(*columnpat.findall(raw_header.get("L", "")))
    header["columns"] = columns
    header["units"] = units

    # extra information like title, angle, etc.
    commentpat = re.compile(r"(?P<name>.*)\s*:\s*(?P<value>.*)\s*\n")
    comments = dict(commentpat.findall(raw_header.get("C", "")))
    header["title"] = comments.get("Title", "")
    header["description"] = comments.get("Notes", "")

    # parse values of the form "Long Name: (value, 'units')" in comments
    valuepat = re.compile(r"[(]\s*(?P<value>.*)\s*, \s*'(?P<units>.*)'\s*[)]")

    def parse_value(valstr):
        d = valuepat.match(valstr).groupdict()
        return float(d["value"]), d["units"]

    if "Detector Angle" in comments:
        header["angle"], _ = parse_value(comments["Detector Angle"])

    return header, data


def write_file(filename, probe, original=None, date=None, title=None, notes=None, run=None, charge=None):
    """
    Save probe as SNS reduced file.
    """
    ## Example header
    # F /SNSlocal/REF_L/2007_1_4B_SCI/2895/NeXus/REF_L_2895.nxs
    # E 1174593434.7
    # D 2007-03-22 15:57:14
    # C Run Number: 2895
    # C Title: MK NR4 dry 032007_No2Rep0
    # C Notes: MK NR 4 DU 53 dry from air
    # C Detector Angle: (0.0, 'degree')
    # C Proton Charge: 45.3205833435

    # S 1 Spectrum ID ('bank1', (85, 151))
    # N 3
    # L time_of_flight(microsecond) data() Sigma()
    from datetime import datetime as dt

    parts = []
    if original is None:
        original = filename
    if date is None:
        date = dt.strftime(dt.now(), "%Y-%m-%d %H:%M:%S")
    parts.append("#F " + original)
    parts.append("#D " + date)
    if run is not None:
        parts.append("#C Run Number: %s" % run)
    if title is not None:
        parts.append("#C Title: %s" % title)
    if notes is not None:
        parts.append("#C Notes: %s" % notes)
    parts.append("#C Detector Angle: (%g, 'degree')" % probe.T[0])
    if charge is not None:
        parts.append("#C Proton Charge: %s" % charge)
    parts.append("")
    parts.append("#N 5")
    parts.append("#L Q(1/A) dQ(1/A) R() dR() L(A)")
    parts.append("")
    header = "\n".join(parts)
    probe.write_data(filename, columns=["Q", "dQ", "R", "dR", "L"], header=header)


class SNSData(object):
    def load(self, filename, **kw):
        return load(filename, instrument=self, **kw)


# TODO: print "Insert correct slit distances for Liquids and Magnetic"
class Liquids(SNSData, Pulsed):
    """
    Loader for reduced data from the SNS Liquids instrument.
    """

    instrument = "Liquids"
    radiation = "neutron"
    feather = LIQUIDS_FEATHER
    wavelength = 2.0, 15.0
    # wavelength = 0.5, 5
    # wavelength = 5.5, 10
    # wavelength = 10.5, 15
    dLoL = 0.02
    d_s1 = 230.0 + 1856.0
    d_s2 = 230.0
    d_moderator = 14.850  # moderator to detector distance
    TOF_range = (6000, 60000)


class Magnetic(SNSData, Pulsed):
    """
    Loader for reduced data from the SNS Magnetic instrument.
    """

    instrument = "Magnetic"
    radiation = "neutron"
    wavelength = 1.8, 14
    dLoL = 0.02
    d_s1 = 75 * 2.54
    d_s2 = 14 * 2.54


# Instrument names assigned by reflpak
INSTRUMENTS = {
    "Liquids": Liquids,
    "Magnetic": Magnetic,
}

# ===== utils ==============


def intensity_from_spline(Lrange, dLoL, feather):
    L0, L1 = Lrange
    n = math.ceil(math.log(L1 / L0) / math.log(1 + dLoL))
    L = L0 * (1 + dLoL) ** np.arange(0, n)
    return (L[:-1] + L[1:]) / 2, rebin(feather[0], feather[1], L)


def boltzmann_feather(L, counts=100000, range=None):
    """
    Return expected intensity as a function of wavelength given the TOF
    feather range and the total number of counts.

    TOF feather is approximately a boltzmann distribution with gaussian
    convolution.  The following looks pretty enough; don't know how well it
    corresponds to the actual SNS feather.
    """
    import scipy.stats

    y = np.linspace(-4, 4, 10)
    G = np.exp(-(y**2) / 10)
    x = np.arange(12, 85)
    B = scipy.stats.boltzmann.pmf(x, 0.05, counts, loc=16)
    BGz = np.convolve(B, G, mode="same")
    # if range is None: range = L[0], L[-1]
    # if range[0] > range[1]: range = range[::-1]
    # range = range[0]*(1-1e-15), range[1]*(1+1e-15)
    # z = np.linspace(range[0], range[1], len(BGz))
    z = np.linspace(2, 16.5, len(BGz))  # Wavelength range for liquids
    pL = np.interp(L, z, BGz, left=0, right=0)
    nL = pL / sum(pL) * counts
    return nL



# pylint: disable=invalid-name
"""
Experiment definition

An experiment combines the sample definition with a measurement probe
to create a fittable reflectometry model.
"""

import json
import os
import traceback
from dataclasses import dataclass
from math import floor, log10, pi
from typing import List, Literal, Optional, Protocol, TypedDict, Union
from warnings import warn

import numpy as np
from bumps import parameter
from bumps.dream.state import MCMCDraw
from bumps.fitproblem import Fitness, FitProblem
from bumps.parameter import Parameter, tag_all

from . import __version__
from .sample.reflectivity import (
    BASE_GUIDE_ANGLE as DEFAULT_THETA_M,
    magnetic_amplitude as reflmag,
    reflectivity_amplitude as reflamp,
)
from . import profile
from .probe.probe import PolarizedNeutronProbe, Probe, QProbe, PolarizedQProbe
from .sample import layers, material
from .utils import asbytes


class WebviewPlotFunction(Protocol):
    def __call__(
        self, model: "ExperimentBase", problem: FitProblem, state: MCMCDraw, n_samples: Optional[int]
    ) -> dict: ...


class WebviewPlotInfo(TypedDict):
    change_with: Literal["parameter", "uncertainty"]
    func: WebviewPlotFunction


class ExperimentBase:
    probe = None  # type: Optional[Probe]
    interpolation = 0
    _probe_cache = None
    _substrate = None
    _surface = None
    _webview_plots: dict[str, WebviewPlotInfo]

    def parameters(self):
        raise NotImplementedError()

    def reflectivity(self, resolution=True, interpolation=0):
        raise NotImplementedError()

    def magnetic_step_profile(self):
        raise NotImplementedError()

    def slabs(self):
        raise NotImplementedError()

    def magnetic_slabs(self):
        raise NotImplementedError()

    def step_profile(self):
        raise NotImplementedError()

    def smooth_profile(self, dz=0.1):
        raise NotImplementedError()

    def plot_profile(self, plot_shift=0.0):
        raise NotImplementedError()

    def format_parameters(self):
        p = self.parameters()
        print(parameter.format(p))

    def update_composition(self):
        """
        When the model composition has changed, we need to lookup the
        scattering factors for the new model.  This is only needed
        when an existing chemical formula is modified; new and
        deleted formulas will be handled automatically.
        """
        self._probe_cache.reset()
        self.update()

    def is_reset(self):
        """
        Returns True if a model reset was triggered.
        """
        return self._cache == {}

    def update(self):
        """
        Called when any parameter in the model is changed.

        This signals that the entire model needs to be recalculated.
        """
        # if we wanted to be particularly clever we could predefine
        # the optical matrices and only adjust those that have changed
        # as the result of a parameter changing.   More trouble than it
        # is worth, methinks.
        # print("reseting calculation")
        self._cache = {}

    def residuals(self):
        if "residuals" not in self._cache:
            # Trigger reflectivity calculation even if there is no data to
            # compare against so that we can profile simulation code, and
            # so that simulation smoke tests are run more thoroughly.
            QR = self.reflectivity()
            if (self.probe.polarized and all(x is None or x.R is None for x in self.probe.xs)) or (
                not self.probe.polarized and self.probe.R is None
            ):
                resid = np.zeros(0)
            else:
                if self.probe.polarized:
                    resid = np.hstack([(xs.R - QRi[1]) / xs.dR for xs, QRi in zip(self.probe.xs, QR) if xs is not None])
                else:
                    resid = (self.probe.R - QR[1]) / self.probe.dR
            self._cache["residuals"] = resid
            # print(("%12s "*4)%("Q", "R", "dR", "Rtheory"))
            # print("\n".join(("%12.6e "*4)%el for el in zip(QR[0], self.probe.R, self.probe.dR, QR[1]))
            # print("resid", np.sum(resid**2)/2)

        return self._cache["residuals"]

    def numpoints(self):
        if self.probe.polarized:
            return sum(len(xs.Q) for xs in self.probe.xs if xs is not None)
        else:
            return len(self.probe.Q) if self.probe.Q is not None else 0

    def nllf(self):
        """
        Return the -log(P(data|model)).

        Using the assumption that data uncertainty is uncorrelated, with
        measurements normally distributed with mean R and variance dR**2,
        this is just sum( resid**2/2 + log(2*pi*dR**2)/2 ).

        The current version drops the constant term, sum(log(2*pi*dR**2)/2).
        """
        # if 'nllf_scale' not in self._cache:
        #    if self.probe.dR is None:
        #        raise ValueError("No data from which to calculate nllf")
        #    self._cache['nllf_scale'] = np.sum(np.log(2*pi*self.probe.dR**2))
        # TODO: add sigma^2 effects back into nllf; only needs to be calculated
        # when dR changes, so maybe it belongs in probe.
        return 0.5 * np.sum(self.residuals() ** 2)  # + self._cache['nllf_scale']

    def plot_reflectivity(self, show_resolution=False, view=None, plot_shift=None):
        n = self.interpolation
        QR = self.reflectivity(interpolation=n)
        self.probe.plot(
            theory=QR,
            substrate=self._substrate,
            surface=self._surface,
            view=view,
            plot_shift=plot_shift,
            interpolation=n,
        )

        if show_resolution:
            import matplotlib.pyplot as plt

            QR = self.reflectivity(resolution=False, interpolation=n)
            if self.probe.polarized:
                # Should be four pairs
                for Q, R in QR:
                    plt.plot(Q, R, ":g")
            else:
                Q, R = QR
                plt.plot(Q, R, ":g")

    def plot(self, plot_shift=None, profile_shift=None, view=None):
        import matplotlib.pyplot as plt

        plt.subplot(211)
        self.plot_reflectivity(plot_shift=plot_shift, view=view)
        plt.subplot(212)
        self.plot_profile(plot_shift=profile_shift)

    def resynth_data(self):
        """Resynthesize data with noise from the uncertainty estimates."""
        self.probe.resynth_data()

    def restore_data(self):
        """Restore original data after resynthesis."""
        self.probe.restore_data()

    def write_data(self, filename, **kw):
        """Save simulated data to a file"""
        self.probe.write_data(filename, **kw)

    def simulate_data(self, noise=2.0):
        """
        Simulate a random data set for the model.

        This sets R and dR according to the noise level given.

        **Parameters:**

        *noise*: float or array or None | %
            dR/R uncertainty as a percentage.  If noise is set to None, then
            use dR from the data if present, otherwise default to 2%.
        """
        # TODO: can't do perfect data while setting default uncertainty
        theory = self.reflectivity(resolution=True)
        self.probe.simulate_data(theory, noise=noise)

    def save(self, basename):
        self.save_profile(basename)
        # self.save_staj(basename)
        self.save_refl(basename)
        self.save_json(basename)

    def save_json(self, basename):
        """Save the experiment as a json file"""
        from bumps.serialize import serialize

        try:
            experiment = serialize(self)
            json_file = basename + "-expt.json"
            with open(json_file, "w") as fid:
                data = json.dumps(experiment)
                fid.write(data)
        except Exception:
            traceback.print_exc()
            warn("failed to create json structure for model")

    def save_profile(self, basename):
        if self.ismagnetic:
            self._save_magnetic(basename)
        else:
            self._save_nonmagnetic(basename)

    def _save_magnetic(self, basename):
        # Slabs
        A = np.array(self.magnetic_slabs())
        header = "# %17s %20s %20s %20s %20s %20s\n" % (
            "thickness (A)",
            "interface (A)",
            "rho (1e-6/A^2)",
            "irho (1e-6/A^2)",
            "rhoM (1e-6/A^2)",
            "theta (degrees)",
        )
        with open(basename + "-slabs.dat", "wb") as fid:
            fid.write(asbytes(header))
            np.savetxt(fid, A.T, fmt="%20.15g")

        # Step profile
        A = np.array(self.magnetic_step_profile())
        header = "# %10s %12s %12s %12s %12s\n" % (
            "z (A)",
            "rho (1e-6/A2)",
            "irho (1e-6/A2)",
            "rhoM (1e-6/A2)",
            "theta (degrees)",
        )
        with open(basename + "-steps.dat", "wb") as fid:
            fid.write(asbytes(header))
            np.savetxt(fid, A.T, fmt="%12.8f")

        # Smooth profile
        A = np.array(self.magnetic_smooth_profile())
        header = "# %10s %12s %12s %12s %12s\n" % (
            "z (A)",
            "rho (1e-6/A2)",
            "irho (1e-6/A2)",
            "rhoM (1e-6/A2)",
            "theta (degrees)",
        )
        with open(basename + "-profile.dat", "wb") as fid:
            fid.write(asbytes(header))
            np.savetxt(fid, A.T, fmt="%12.8f")

    def _save_nonmagnetic(self, basename):
        # Slabs
        A = np.array(self.slabs())
        header = "# %17s %20s %20s %20s\n" % ("thickness (A)", "interface (A)", "rho (1e-6/A^2)", "irho (1e-6/A^2)")
        with open(basename + "-slabs.dat", "wb") as fid:
            fid.write(asbytes(header))
            np.savetxt(fid, A.T, fmt="%20.15g")

        # Step profile
        A = np.array(self.step_profile())
        header = "# %10s %20s %20s\n" % ("z (A)", "rho (1e-6/A2)", "irho (1e-6/A2)")
        with open(basename + "-steps.dat", "wb") as fid:
            fid.write(asbytes(header))
            np.savetxt(fid, A.T, fmt="%12.8f")

        # Smooth profile
        A = np.array(self.smooth_profile())
        header = "# %10s %12s %12s\n" % ("z (A)", "rho (1e-6/A2)", "irho (1e-6/A2)")
        with open(basename + "-profile.dat", "wb") as fid:
            fid.write(asbytes(header))
            np.savetxt(fid, A.T, fmt="%12.8f")

    def save_refl(self, basename):
        # Reflectivity
        theory = self.reflectivity()
        self.probe.save(
            filename=basename + "-refl.dat", theory=theory, substrate=self._substrate, surface=self._surface
        )
        if self.interpolation > 0:
            theory = self.reflectivity(interpolation=self.interpolation)
            self.probe.save(filename=basename + "-refl-interp.dat", theory=theory)

    def register_webview_plot(
        self, plot_title: str, plot_function: WebviewPlotFunction, change_with: Literal["parameter", "uncertainty"]
    ):
        # change_with = 'parameter' or 'uncertainty'
        # Plot function syntax: f(model, problem) for 'parameter'
        #                       f(model, problem, state, n_samples) for 'uncertainty'

        self._webview_plots[plot_title] = dict(change_with=change_with, func=plot_function)

    @property
    def webview_plots(self):
        return self._webview_plots


@dataclass(init=False)
class Experiment(ExperimentBase):
    """
    Theory calculator.  Associates sample with data, Sample plus data.
    Associate sample with measurement.

    The model calculator is specific to the particular measurement technique
    that was applied to the model.

    Measurement properties:

        *probe* is the measuring probe

    Sample properties:

        *sample* is the model sample
        *step_interfaces* use slabs to approximate gaussian interfaces
        *roughness_limit* limit the roughness based on layer thickness
        *dz* minimum step size for computed profile steps in Angstroms
        *dA* discretization condition for computed profiles

    If *step_interfaces* is True, then approximate the interface using
    microslabs with step size *dz*.  The microslabs extend throughout
    the whole profile, both the interfaces and the bulk; a value
    for *dA* should be specified to save computation time.  If False, then
    use the Nevot-Croce analytic expression for the interface between slabs.

    The *roughness_limit* value should be reasonably large (e.g., 2.5 or above)
    to make sure that the Nevot-Croce reflectivity calculation matches the
    calculation of the displayed profile.  Use a value of 0 if you want no
    limits on the roughness,  but be aware that the displayed profile may
    not reflect the actual scattering densities in the material.

    The *dz* step size sets the size of the slabs for non-uniform profiles.
    Using the relation d = 2 pi / Q_max,  we use a default step size of d/20
    rounded to two digits, with 5 |Ang| as the maximum default.  For
    simultaneous fitting you may want to set *dz* explicitly using to
    round(pi/Q_max/10, 1) so that all models use the same step size.

    The *dA* condition measures the uncertainty in scattering materials
    allowed when combining the steps of a non-uniform profile into slabs.
    Specifically, the area of the box containing the minimum and the
    maximum of the non-uniform profile within the slab will be smaller
    than *dA*.  A *dA* of 10 gives coarse slabs.  If *dA* is not provided
    then each profile step forms its own slab.  The *dA* condition will
    also apply to the slab approximation to the interfaces.

    *interpolation* indicates the number of points to plot in between
    existing points.

    *smoothness* **DEPRECATED** This parameter is not used.
    """

    name: str
    sample: Optional[layers.Stack]
    probe: Union[Probe, PolarizedNeutronProbe, QProbe, PolarizedQProbe]
    roughness_limit: float
    dz: Union[float, Literal[None]]
    dA: Union[float, Literal[None]]
    step_interfaces: bool
    interpolation: float
    version: str

    profile_shift = 0

    def __init__(
        self,
        sample: Optional[layers.Stack] = None,
        probe=None,
        name=None,
        roughness_limit=0,
        dz=None,
        dA=None,
        step_interfaces=None,
        smoothness=None,
        interpolation=0,
        constraints=None,
        version: Optional[str] = None,
        auto_tag=False,
    ):
        # Note: smoothness ignored
        self.sample = sample
        self._substrate = self.sample[0].material
        self._surface = self.sample[-1].material
        self.probe = probe
        self.roughness_limit = roughness_limit
        if dz is None:
            dz = nice((2 * pi / probe.Q.max()) / 10)
            dz = min(dz, 5.0)
        self.dz = dz
        self.dA = dA
        self.step_interfaces = step_interfaces
        self.interpolation = interpolation
        # TODO: proper 2D mesh resolution over L, T
        # (currently R is only calculated for first L anyway)
        # num_slabs = len(probe.unique_L) if probe.unique_L is not None else 1
        num_slabs = 1
        self._slabs = profile.Microslabs(num_slabs, dz=dz)
        self._probe_cache = material.ProbeCache(probe)
        self._cache = {}  # Cache calculated profiles/reflectivities
        self.name = name if name is not None else probe.name
        self.constraints = constraints
        self.version = __version__ if version is None else version
        if auto_tag:
            tag_all(self.probe.parameters(), "probe")
            tag_all(self.sample.parameters(), "sample")
        self._webview_plots = {}

    @property
    def ismagnetic(self):
        """True if experiment contains magnetic materials"""
        slabs = self._render_slabs()
        return slabs.ismagnetic

    def parameters(self):
        """Fittable parameters to sample and probe"""
        return {
            "sample": self.sample.parameters(),
            "probe": self.probe.parameters(),
        }

    def _render_slabs(self):
        """
        Build a slab description of the model from the individual layers.
        """
        key = "rendered", self.step_interfaces, self.dA
        if key not in self._cache:
            self._slabs.clear()
            self.sample.render(self._probe_cache, self._slabs)
            self._slabs.finalize(step_interfaces=self.step_interfaces, dA=self.dA)
            # roughness_limit=self.roughness_limit)
            self._cache[key] = True
        return self._slabs

    def _reflamp(self):
        # calc_q = self.probe.calc_Q
        # return calc_q, calc_q
        key = "calc_r"
        if key not in self._cache:
            slabs = self._render_slabs()
            w = slabs.w
            rho, irho = slabs.rho, slabs.irho
            sigma = slabs.sigma
            # sigma = slabs.sigma
            calc_q = self.probe.calc_Q
            # print("calc Q", self.probe.calc_Q)
            if slabs.ismagnetic:
                rhoM, thetaM = slabs.rhoM, slabs.thetaM
                Aguide = self.probe.Aguide.value
                H = self.probe.H.value
                calc_r = reflmag(
                    -calc_q / 2,
                    depth=w,
                    rho=rho[0],
                    irho=irho[0],
                    rhoM=rhoM,
                    thetaM=thetaM,
                    Aguide=Aguide,
                    H=H,
                    sigma=sigma,
                )
            else:
                calc_r = reflamp(-calc_q / 2, depth=w, rho=rho, irho=irho, sigma=sigma)
            if False and np.isnan(calc_r).any():
                print("w", w)
                print("rho", rho)
                print("irho", irho)
                if slabs.ismagnetic:
                    print("rhoM", rhoM)
                    print("thetaM", thetaM)
                    print("Aguide", Aguide, "H", H)
                print("sigma", sigma)
                print("kz", self.probe.calc_Q / 2)
                print("R", abs(np.asarray(calc_r) ** 2))
                pars = parameter.unique(self.parameters())
                fitted = parameter.varying(pars)
                print(parameter.summarize(fitted))
                print("===")
            self._cache[key] = calc_q, calc_r
            # if np.isnan(calc_q).any(): print("calc_Q contains NaN")
            # if np.isnan(calc_r).any(): print("calc_r contains NaN")
        return self._cache[key]

    def amplitude(self, resolution=False, interpolation=0):
        """
        Calculate reflectivity amplitude at the probe points.
        """
        key = ("amplitude", resolution, interpolation)
        if key not in self._cache:
            calc_q, calc_r = self._reflamp()
            res = self.probe.apply_beam(calc_q, calc_r, resolution=resolution, interpolation=interpolation)
            self._cache[key] = res
        return self._cache[key]

    def reflectivity(self, resolution=True, interpolation=0):
        """
        Calculate predicted reflectivity.

        If *resolution* is true include resolution effects.
        """
        key = ("reflectivity", resolution, interpolation)
        if key not in self._cache:
            calc_q, calc_r = self._reflamp()
            calc_R = _amplitude_to_magnitude(calc_r, ismagnetic=self.ismagnetic, polarized=self.probe.polarized)
            res = self.probe.apply_beam(calc_q, calc_R, resolution=resolution, interpolation=interpolation)
            self._cache[key] = res
        return self._cache[key]

    def smooth_profile(self, dz=0.1):
        """
        Return the scattering potential for the sample.

        If *dz* is not given, use *dz* = 0.1 A.
        """
        if self.step_interfaces:
            return self.step_profile()
        key = "smooth_profile", dz
        if key not in self._cache:
            slabs = self._render_slabs()
            prof = slabs.smooth_profile(dz=dz)
            self._cache[key] = prof
        return self._cache[key]

    def step_profile(self):
        """
        Return the step scattering potential for the sample, ignoring
        interfaces.
        """
        key = "step_profile"
        if key not in self._cache:
            slabs = self._render_slabs()
            prof = slabs.step_profile()
            self._cache[key] = prof
        return self._cache[key]

    def slabs(self):
        """
        Return the slab thickness, roughness, rho, irho for the
        rendered model.

        .. Note::
             Roughness is for the top of the layer.
        """
        slabs = self._render_slabs()
        return (slabs.w, np.hstack((slabs.sigma, 0)), slabs.rho[0], slabs.irho[0])

    def magnetic_smooth_profile(self, dz=0.1):
        """
        Return the nuclear and magnetic scattering potential for the sample.
        """
        key = "magnetic_smooth_profile", "{:.6f}".format(dz)
        if key not in self._cache:
            slabs = self._render_slabs()
            prof = slabs.magnetic_smooth_profile(dz=dz)
            self._cache[key] = prof
        return self._cache[key]

    def magnetic_step_profile(self):
        """
        Return the nuclear and magnetic scattering potential for the sample.
        """
        key = "magnetic_step_profile"
        if key not in self._cache:
            slabs = self._render_slabs()
            prof = slabs.magnetic_step_profile()
            self._cache[key] = prof
        return self._cache[key]

    def magnetic_slabs(self):
        slabs = self._render_slabs()
        return (slabs.w, np.hstack((slabs.sigma, 0)), slabs.rho[0], slabs.irho[0], slabs.rhoM, slabs.thetaM)

    def save_staj(self, basename):
        from .probe.data_loaders.stajconvert import save_mlayer

        try:
            if self.probe.R is not None:
                datafile = getattr(self.probe, "filename", basename + ".refl")
            else:
                datafile = None
            save_mlayer(self, basename + ".staj", datafile=datafile)
            probe = self.probe
            datafile = os.path.join(os.path.dirname(basename), os.path.basename(datafile))
            header = "# Q R dR\n"
            with open(datafile, "wb") as fid:
                fid.write(asbytes(header))
                np.savetxt(fid, np.vstack((probe.Qo, probe.R, probe.dR)).T)
            fid.close()
        except Exception:
            print("==== could not save staj file ====")
            traceback.print_exc()

    def plot_profile(self, plot_shift=None):
        import matplotlib.pyplot as plt
        from bumps.plotutil import auto_shift

        plot_shift = plot_shift if plot_shift is not None else Experiment.profile_shift
        trans = auto_shift(plot_shift)
        if self.ismagnetic:
            if not self.step_interfaces:
                z, rho, irho, rhoM, thetaM = self.magnetic_step_profile()
                # rhoM_net = rhoM*np.cos(np.radians(thetaM))
                plt.plot(z, rho, ":g", transform=trans)
                plt.plot(z, irho, ":b", transform=trans)
                plt.plot(z, rhoM, ":r", transform=trans)
                if (abs(thetaM - DEFAULT_THETA_M) > 1e-3).any():
                    ax = plt.twinx()
                    ax.plot(z, thetaM, ":k", transform=trans)
                    plt.ylabel("magnetic angle (degrees)")
            z, rho, irho, rhoM, thetaM = self.magnetic_smooth_profile()
            # rhoM_net = rhoM*np.cos(np.radians(thetaM))
            handles = [
                plt.plot(z, rho, "-g", transform=trans, label="rho")[0],
                plt.plot(z, irho, "-b", transform=trans, label="irho")[0],
                plt.plot(z, rhoM, "-r", transform=trans, label="rhoM")[0],
            ]
            if (abs(thetaM - DEFAULT_THETA_M) > 1e-3).any():
                ax = plt.twinx()
                h = ax.plot(z, thetaM, "-k", transform=trans, label="thetaM")
                handles.append(h[0])
                plt.ylabel("magnetic angle (degrees)")
            plt.xlabel("depth (A)")
            plt.ylabel("SLD (10^6 / A**2)")
            labels = [h.get_label() for h in handles]
            plt.legend(handles=handles, labels=labels)
        else:
            if not self.step_interfaces:
                z, rho, irho = self.step_profile()
                plt.plot(z, rho, ":g", z, irho, ":b", transform=trans)
            z, rho, irho = self.smooth_profile()
            plt.plot(z, rho, "-g", z, irho, "-b", transform=trans)
            plt.legend(["rho", "irho"])
            plt.xlabel("depth (A)")
            plt.ylabel("SLD (10^6 / A**2)")

    def penalty(self):
        return self.sample.penalty()


assert isinstance(Experiment, Fitness)


@dataclass(init=False)
class MixedExperiment(ExperimentBase):
    """
    Support composite sample reflectivity measurements.

    Sometimes the sample you are measuring is not uniform.
    For example, you may have one portion of you polymer
    brush sample where the brushes are close packed and able
    to stay upright, whereas a different section of the sample
    has the brushes lying flat.  Constructing two sample
    models, one with brushes upright and one with brushes
    flat, and adding the reflectivity incoherently, you can
    then fit the ratio of upright to flat.

    *samples* the layer stacks making up the models
    *ratio* a list of parameters, such as [3, 1] for a 3:1 ratio
    *probe* the measurement to be fitted or simulated

    *coherent* is True if the length scale of the domains
    is less than the coherence length of the neutron, or false
    otherwise.

    Statistics such as the cost functions for the individual
    profiles can be accessed from the underlying experiments
    using composite.parts[i] for the various samples.
    """

    name: str
    ratio: List[Union[float, Parameter]]
    samples: Optional[List[layers.Stack]]
    probe: Union[Probe, PolarizedNeutronProbe]
    coherent: bool
    interpolation: float

    def __init__(self, samples=None, ratio=None, probe=None, name=None, coherent=False, interpolation=0, **kw):
        self.samples = samples
        self.probe = probe
        self.ratio = [Parameter.default(r, name="ratio %d" % i) for i, r in enumerate(ratio)]
        self.parts = [Experiment(s, probe, name=s.name, **kw) for s in samples]
        self.coherent = coherent
        self.interpolation = interpolation
        self._substrate = self.samples[0][0].material
        self._surface = self.samples[0][-1].material
        self._cache = {}
        self.name = name if name is not None else probe.name
        self._webview_plots = {}

    def update(self):
        self._cache = {}
        for p in self.parts:
            p.update()

    def parameters(self):
        return {
            "samples": [s.parameters() for s in self.samples],
            "ratio": self.ratio,
            "probe": self.probe.parameters(),
        }

    def _reflamp(self):
        """
        Calculate the amplitude of the reflectivity...

        For an incoherent sum, we want to add the squares of the amplitudes,
        with a weighting specified by self.ratio, so the amplitudes
        are scaled by sqrt(self.ratio/total) so when they get squared and added
        the normalization is correct.

        For a coherent sum, just multiply by ratio/total.
        It all comes out in the wash.
        """
        total = sum(r.value for r in self.ratio)
        Qs, Rs = zip(*[p._reflamp() for p in self.parts])
        if not self.coherent:
            Rs = [np.asarray(ri) * np.sqrt(ratio_i.value / total) for ri, ratio_i in zip(Rs, self.ratio)]
        else:  # self.coherent == True
            Rs = [np.asarray(ri) * (ratio_i.value / total) for ri, ratio_i in zip(Rs, self.ratio)]
        # print("Rs", Rs)
        return Qs[0], Rs

    def amplitude(self, resolution=False):
        """ """
        if not self.coherent:
            raise TypeError("Cannot compute amplitude of system which is mixed incoherently")
        key = ("amplitude", resolution)
        if key not in self._cache:
            calc_Q, calc_R = self._reflamp()
            calc_R = np.sum(calc_R, axis=1)
            r_real = self.probe.apply_beam(calc_Q, calc_R.real, resolution=resolution)
            r_imag = self.probe.apply_beam(calc_Q, calc_R.imag, resolution=resolution)
            r = r_real + 1j * r_imag
            self._cache[key] = self.probe.Q, r
        return self._cache[key]

    def reflectivity(self, resolution=True, interpolation=0):
        """
        Calculate predicted reflectivity.

        This will be the weighted sum of the reflectivity from the
        individual systems.  If coherent is set, then the coherent
        sum will be used, otherwise the incoherent sum will be used.

        If *resolution* is true include resolution effects.

        *interpolation* is the number of theory points to show between data
        points.
        """
        key = ("reflectivity", resolution, interpolation)
        if key not in self._cache:
            Q, r = self._reflamp()

            polarized = self.probe.polarized
            ismagnetic = any(p.ismagnetic for p in self.parts)

            # If any reflectivity is magnetic, make all reflectivity magnetic
            if ismagnetic:
                for i, p in enumerate(self.parts):
                    if not p.ismagnetic:
                        r[i] = _polarized_nonmagnetic(r[i])

            # Add the cross sections
            if self.coherent:
                r = np.sum(r, axis=0)
                R = _amplitude_to_magnitude(r, ismagnetic=ismagnetic, polarized=polarized)
            else:
                R = [_amplitude_to_magnitude(ri, ismagnetic=ismagnetic, polarized=polarized) for ri in r]
                R = np.sum(R, axis=0)

            # Apply resolution
            res = self.probe.apply_beam(Q, R, resolution=resolution, interpolation=0)
            self._cache[key] = res
        return self._cache[key]

    def plot_profile(self, plot_shift=None):
        f = np.array([r.value for r in self.ratio], "d")
        f /= np.sum(f)
        for p in self.parts:
            p.plot_profile(plot_shift=plot_shift)

    def save_profile(self, basename):
        for i, p in enumerate(self.parts):
            p.save_profile("%s-%d" % (basename, i))

    def save_staj(self, basename):
        for i, p in enumerate(self.parts):
            p.save_staj("%s-%d" % (basename, i))

    def penalty(self):
        return sum(s.penalty() for s in self.samples)


def nice(v, digits=2):
    """Fix v to a value with a given number of digits of precision"""
    if v == 0.0:
        return v
    sign = v / abs(v)
    place = floor(log10(abs(v)))
    scale = 10 ** (place - (digits - 1))
    return sign * floor(abs(v) / scale + 0.5) * scale


def plot_sample(sample, instrument=None, roughness_limit=0):
    """
    Quick plot of a reflectivity sample and the corresponding reflectivity.
    """
    if instrument is None:
        from .probe.probe import NeutronProbe

        probe = NeutronProbe(T=np.arange(0, 5, 0.05), L=5)
    else:
        probe = instrument.simulate()
    experiment = Experiment(sample=sample, probe=probe, roughness_limit=roughness_limit)
    experiment.plot()


def _polarized_nonmagnetic(r):
    """Convert nonmagnetic data to polarized representation.

    Polarized non-magnetic data repeats the reflectivity in the non spin flip
    channels and sets the spin flip channels to zero.
    """
    nsf = r
    sf = 0 * r
    return [nsf, sf, sf, nsf]


def _nonpolarized_magnetic(R):
    """Convert magnetic reflectivity to unpolarized representation.

    Unpolarized magnetic data adds the cross-sections of the magnetic
    data incoherently and divides by two.
    """
    return sum(R) / 2


def _amplitude_to_magnitude(r, ismagnetic, polarized):
    """
    Compute the reflectivity magnitude
    """
    if ismagnetic:
        R = [abs(xs) ** 2 for xs in r]
        if not polarized:
            R = _nonpolarized_magnetic(R)
    else:
        R = abs(r) ** 2
        if polarized:
            R = _polarized_nonmagnetic(R)
    return R



from typing import TYPE_CHECKING, Optional

import numpy as np

if TYPE_CHECKING:
    import plotly.graph_objs as go

from refl1d.uncertainty import _find_offset, align_profiles, form_quantiles
from .colors import COLORS

ErrorType = tuple[
    dict[str, list[np.ndarray]],  # profiles
    dict[str, np.ndarray],  # slabs
    dict[str, np.ndarray],  # Q
    dict[str, np.ndarray],  # residuals
]

CONTOURS = (68, 95)


def get_color(index: int):
    return COLORS[index % len(COLORS)]


def show_errors(errors: ErrorType, npoints: int, align: bool, residuals: bool = True):
    from plotly.subplots import make_subplots

    specs = (
        [[{"secondary_y": True, "rowspan": 1}], [{}]] if residuals else [[{"secondary_y": True, "rowspan": 2}], [None]]
    )
    fig = make_subplots(rows=2, cols=1, specs=specs)
    fig.update_layout(template=None)
    contour_data = show_profiles(errors, align, CONTOURS, npoints, fig=fig, row=1, col=1)
    if residuals:
        show_residuals(errors, None, fig=fig, row=2, col=1)
    return dict(fig=fig, contour_data=contour_data, contours=CONTOURS)


def show_profiles(
    errors: ErrorType,
    align: bool,
    contours: Optional[tuple[float]],
    npoints: int,
    fig: "go.Figure",
    row: Optional[int] = None,
    col: Optional[int] = None,
) -> "go.Figure":
    profiles, slabs, _, _ = errors
    if align is not None:
        profiles = align_profiles(profiles, slabs, align)

    contour_data = None
    if contours is not None:
        contour_data = _profiles_contour(profiles, contours, npoints, fig=fig, row=row, col=col)
    else:
        _profiles_overplot(profiles, fig=fig, row=row, col=col)

    if align != "auto":
        _profiles_draw_align_lines(profiles, slabs, align, fig=fig, row=row, col=col)

    return contour_data


def show_residuals(
    errors: ErrorType,
    contours: Optional[tuple[float]],
    fig: "go.Figure",
    row: Optional[int] = None,
    col: Optional[int] = None,
):
    _, _, Q, residuals = errors

    if contours is not None:
        _residuals_contour(Q, residuals, contours, fig, row, col)
    else:
        _residuals_overplot(Q, residuals, fig=fig, row=row, col=col)


def _profiles_overplot(
    profiles: dict[str, list[np.ndarray]], fig: "go.Figure", row: Optional[int] = None, col: Optional[int] = None
):
    has_magnetism = False
    for model_index, (model, group) in enumerate(profiles.items()):
        name = model.name
        absorbing = any((L[2] != 1e-4).any() for L in group)
        magnetic = len(group[0]) > 3
        # Note: Use 3 colours per dataset for consistency
        color_index = model_index * 4
        _draw_overplot(group, 1, name + " rho", fig=fig, color_index=color_index, row=row, col=col, secondary_y=False)
        if absorbing:
            _draw_overplot(
                group, 2, name + " irho", fig=fig, color_index=color_index + 1, row=row, col=col, secondary_y=False
            )
        if magnetic:
            has_magnetism = True
            _draw_overplot(
                group, 3, name + " rhoM", fig=fig, color_index=color_index + 2, row=row, col=col, secondary_y=False
            )
            _draw_overplot(
                group, 4, name + " thetaM", fig=fig, color_index=color_index + 3, row=row, col=col, secondary_y=True
            )
    _profile_labels(fig=fig, row=row, col=col, magnetic=has_magnetism)


def _draw_overplot(
    group,
    index,
    label,
    fig: "go.Figure",
    color_index: int,
    row: Optional[int] = None,
    col: Optional[int] = None,
    secondary_y: bool = False,
):
    color = get_color(color_index)
    for L in group[1:]:
        fig.add_scattergl(
            x=L[0],
            y=L[index],
            opacity=0.1,
            showlegend=False,
            mode="lines",
            line={"color": color},
            hoverinfo="skip",
            row=row,
            col=col,
            secondary_y=secondary_y,
        )
    # Plot best
    L = group[0]
    fig.add_scattergl(
        x=L[0],
        y=L[index],
        name=label,
        mode="lines",
        line={"color": color},
        hoverinfo="skip",
        row=row,
        col=col,
        secondary_y=secondary_y,
    )


def _profiles_contour(
    profiles: dict[str, list[np.ndarray]],
    contours: Optional[tuple[float]],
    npoints,
    fig: "go.Figure",
    row: Optional[int] = None,
    col: Optional[int] = None,
):
    contour_data = {}
    has_magnetism = False
    for model_index, (model, group) in enumerate(profiles.items()):
        name = model.name if model.name is not None else f"Model {model_index}"
        if name in contour_data:
            name += f" {model_index}"
        contour_data[name] = {}
        absorbing = any((L[2] > 1e-4).any() for L in group)
        magnetic = len(group[0]) > 3
        # Find limits of all profiles
        z = np.hstack([line[0] for line in group])
        zp = np.linspace(np.min(z), np.max(z), npoints)
        contour_data[name]["z"] = zp
        contour_data[name]["data"] = {}
        # Note: Use 4 colours per dataset for consistency
        color_index = model_index * 4
        contour_data[name]["data"]["rho"] = _draw_contours(
            group, 1, name + " rho", zp, contours, fig=fig, color_index=color_index, row=row, col=col, secondary_y=False
        )
        if absorbing:
            contour_data[name]["data"]["irho"] = _draw_contours(
                group,
                2,
                name + " irho",
                zp,
                contours,
                fig=fig,
                color_index=color_index + 1,
                row=row,
                col=col,
                secondary_y=False,
            )
        if magnetic:
            has_magnetism = True
            contour_data[name]["data"]["rhoM"] = _draw_contours(
                group,
                3,
                name + " rhoM",
                zp,
                contours,
                fig=fig,
                color_index=color_index + 2,
                row=row,
                col=col,
                secondary_y=False,
            )
            contour_data[name]["data"]["thetaM"] = _draw_contours(
                group,
                4,
                name + " thetaM",
                zp,
                contours,
                fig=fig,
                color_index=color_index + 3,
                row=row,
                col=col,
                secondary_y=True,
            )

    _profile_labels(fig=fig, row=row, col=col, magnetic=has_magnetism)
    return contour_data


def _draw_contours(
    group,
    index,
    label,
    zp,
    contours,
    fig: "go.Figure",
    color_index: int,
    row: Optional[int] = None,
    col: Optional[int] = None,
    secondary_y: bool = False,
):
    # Interpolate on common z
    fp = np.vstack([np.interp(zp, L[0], L[index]) for L in group])
    # Plot the quantiles
    color = get_color(color_index)
    legendgroup = f"group_{color_index}"
    fig.add_scattergl(
        x=zp,
        y=fp[0],
        mode="lines",
        name=label,
        line=dict(color=color),
        legendgroup=legendgroup,
        row=row,
        col=col,
        secondary_y=secondary_y,
    )
    named_contours = _plot_quantiles(
        zp, fp, contours, color, alpha=None, fig=fig, row=row, col=col, legendgroup=legendgroup, secondary_y=secondary_y
    )
    named_contours["best"] = fp[0]
    # Plot the best
    # axes.plot(zp, fp[0], '-', label=label, color=dark(color))
    return named_contours


def _profile_labels(fig: "go.Figure", row: Optional[int] = None, col: Optional[int] = None, magnetic: bool = False):
    fig.update_xaxes(title_text="z (Å)", row=row, col=col, showline=True, zeroline=False)
    fig.update_yaxes(title_text="SLD (10⁻⁶/Å²)", row=row, col=col, showline=True, secondary_y=False, side="left")
    if magnetic:
        fig.update_yaxes(
            automargin=True,
            showgrid=False,
            title_text="Magnetic Angle θ<sub>M</sub> / °",
            row=row,
            col=col,
            showline=True,
            secondary_y=True,
            side="right",
        )


def _residuals_overplot(Q, residuals, fig: "go.Figure", row: Optional[int] = None, col: Optional[int] = None):
    alpha = 0.6
    shift = 0
    for m_index, (m, r) in enumerate(residuals.items()):
        color_index = m_index * 4
        color = get_color(color_index)
        pop_size = r.shape[1] - 1
        qq = np.tile(Q[m], pop_size)
        rr = shift + r[:, 1:].ravel(order="F")
        fig.add_scattergl(
            x=qq,
            y=rr,
            showlegend=False,
            mode="markers",
            marker=dict(color=color, size=2),
            opacity=alpha,
            row=row,
            col=col,
            hoverinfo="skip",
        )
        fig.add_scattergl(
            x=Q[m],
            y=shift + r[:, 0],
            name=m.name,
            mode="markers",
            marker=dict(color=color, size=3),
            row=row,
            col=col,
            hoverinfo="skip",
        )
        shift += 5
    _residuals_labels(fig, row=row, col=col)


def _residuals_contour(Q, residuals, contours, fig: "go.Figure", row: Optional[int] = None, col: Optional[int] = None):
    shift = 0
    for model_index, (m, r) in enumerate(residuals.items()):
        color_index = model_index * 4
        color = get_color(color_index)
        x = Q[m]
        # residuals x may not be sorted:
        sort_order = np.argsort(x)
        sorted_x = x[sort_order]
        fig.add_scattergl(
            x=sorted_x,
            y=shift + r[:, 0],
            mode="markers",
            name=m.name,
            marker=dict(color=color, size=3),
            row=row,
            col=col,
        )
        _plot_quantiles(sorted_x, shift + r.T[:, sort_order], contours, color, alpha=None, fig=fig, row=row, col=col)
        # Plot the best
        shift += 5
    _residuals_labels(fig, row=row, col=col)


def _residuals_labels(fig, row=None, col=None):
    fig.update_xaxes(title_text="Q (1/Å)", row=row, col=col, showline=True, zeroline=False)
    fig.update_yaxes(title_text="Residuals", row=row, col=col, showline=True)


def _profiles_draw_align_lines(
    profiles, slabs, align, fig: "go.Figure", row: Optional[int] = None, col: Optional[int] = None
):
    for i, m in enumerate(profiles.keys()):
        t1_offset = _find_offset(slabs[m][0], align) if align != "auto" else None
        if t1_offset is not None:
            fig.add_vline(x=t1_offset, line=dict(dash="dash"), row=row, col=col)


def _plot_quantiles(
    x,
    y,
    contours,
    color,
    alpha,
    fig: "go.Figure",
    row: Optional[int] = None,
    col: Optional[int] = None,
    legendgroup: Optional[str] = None,
    secondary_y: bool = False,
):
    """
    Plot quantile curves for a set of lines.

    *x* is the x coordinates for all lines.

    *y* is the y coordinates, one row for each line.

    *contours* is a list of confidence intervals expressed as percents.

    *color* is the color to use for the quantiles.  Quantiles are draw as
    a filled region with alpha transparency.  Higher probability regions
    will be covered with multiple contours, which will make them lighter
    and more saturated.

    *alpha* is the transparency level to use for all fill regions.  The
    default value, alpha=2./(#contours+1), works pretty well.
    """
    _, q = form_quantiles(y, contours)

    output = {}
    if alpha is None:
        alpha = 2.0 / (len(q) + 1)
    for contour, (lo, hi) in zip(contours, q):
        output[f"{contour} percent lower"] = lo
        output[f"{contour} percent upper"] = hi
        fig.add_scattergl(
            x=x,
            y=lo,
            showlegend=False,
            mode="lines",
            line=dict(color=color, width=1),
            opacity=alpha,
            row=row,
            col=col,
            hoverinfo="skip",
            legendgroup=legendgroup,
            secondary_y=secondary_y,
        )
        fig.add_scattergl(
            x=x,
            y=hi,
            showlegend=False,
            mode="lines",
            line=dict(color=color, width=1),
            fill="tonexty",
            opacity=alpha,
            row=row,
            col=col,
            hoverinfo="skip",
            legendgroup=legendgroup,
            secondary_y=secondary_y,
        )
    return output



def convolve_point_sampled(Nin, xin, yin, Np, xp, yp, xo, dx, _in):
    # Walk the theory spline and the resolution spline together, computing
    # the integral of the pairs of line segments.  Since the spline knots
    # do not necessarily correspond, we need to treat each individual segment
    # in both curves piece-wise, integrating from knot to knot in the union
    # of the two knot sets.

    # Need an undefined value in case the algorithm is broken and we don't
    # initialize the initial lines.  This would ideal be NaN, but that isn't
    # available in MSVC, so set it to something very large.  In order to avoid
    # normalizing that very large value to 1., only set the theory line to
    # undefined, and set the resolution line to 0.
    # const double undefined = 0./0.;
    undefined = 1e308
    m1 = undefined
    b1 = undefined
    m2 = 0.0
    b2 = 0.0
    sum = 0.0
    norm = 0.0
    # size_t p;
    # double delta,delta2,delta3;
    # double x, next_x, next_xin, next_xp;

    # Set the target start of the integral to xin[_in].  This may or may
    # not be beyond the end of the resolution function, depending on whether
    # we are at the left edge of the data, or somewhere inside.
    next_xin = xin[_in]

    # Set p to the point just before next_xin
    for p in range(1, Np):
        if xo + dx * xp[p] > next_xin:
            break

    p -= 1
    next_xp = xo + dx * xp[p]

    # Choose the larger of next_xp and next_xin as the starting point of the
    # integral.  This will force us to step both xin and xp in the first
    # iteration of the loop, computing new slope/intercepts for both lines.
    # If the theory line extends beyond the resolution function, then we will
    # be called with xin[_in] before the first resolution point, and next_xp
    # will be bigger than next_xin.  If the resolution extends beyond the
    # end of the theory, then the next_xp will be set before next_xin, and
    # next_xin will be the start of the integral.  The integral ends when
    # either the resolution or the theory runs out.
    #
    # We are tracking the area under the resolution as well as the area
    # under the product of theory and resolution so that we can properly
    # normalize the data when less than the full resolution is included.
    # This means that we do not need to provide a resolution function
    # normalized to a total area of 1 as our input.
    x = next_xp if next_xp > next_xin else next_xin
    # printf("  point xo:%g dx:%g x:%g, p:%ld, in:%ld, xp:%g, xin:%g\n",
    #       xo,dx,x,p,in,next_xp,next_xin);
    while True:
        # Step xin if we are at the next theory point
        if next_xin <= x:
            _in += 1
            if _in >= Nin:
                break  # At the right edge of the data
            next_xin = xin[_in]
            m1 = (yin[_in] - yin[_in - 1]) / (xin[_in] - xin[_in - 1])
            b1 = yin[_in] - m1 * xin[_in]

        # Step xp if we are at the next resolution point
        if next_xp <= x:
            p += 1
            if p >= Np:
                break  # At the right edge of the resolution
            next_xp = xo + dx * xp[p]
            m2 = (yp[p] - yp[p - 1]) / (xp[p] - xp[p - 1]) / dx
            b2 = yp[p] - m2 * next_xp

        # Find the next node
        next_x = next_xin if next_xin < next_xp else next_xp
        # Compute the convolution and norm between the current and next node
        delta = next_x - x
        delta2 = next_x * next_x - x * x
        delta3 = next_x * next_x * next_x - x * x * x
        norm += 0.5 * m2 * delta2 + b2 * delta
        sum += m1 * m2 / 3.0 * delta3 + 0.5 * (m1 * b2 + m2 * b1) * delta2 + b1 * b2 * delta
        # printf("  delta:%g delta2:%g delta3:%g norm:%g sum:%g m1:%g b1:%g m2:%g b2:%g x:%g nx:%g ni:%g np:%g\n",
        #  delta, delta2, delta3, norm, sum, m1, b1, m2, b2, x, next_x, next_xin, next_xp);
        # Move to the next node
        x = next_x

    return sum / norm


def convolve_sampled(xin, yin, xp, yp, x, dx, y):
    Nin = len(xin)
    Np = len(xp)
    N = len(x)

    # /* FIXME fails if xin are not sorted
    # * slow if x not sorted */
    if not Nin > 1:
        raise ValueError("Nin must be > 1")

    # /* Scan through all x values to be calculated */
    # /* Re: omp, each thread is going through the entire input array,
    # * independently, computing the resolution from the neighbourhood
    # * around its individual output points.  The firstprivate( in )
    # * clause sets each thread to keep its own copy of in, initialized
    # * at in's initial value of zero.  The "schedule(static,1)" clause
    # * puts neighbouring points in separate threads, which is a benefit
    # * since there will be less backtracking if resolution width increases
    # * from point to point.  Because the schedule is static, this does not
    # * significantly increase the parallelization overhead.  Because the
    # * threads are operating on interleaved points, there should be fewer cache
    # * misses than if each thread were given different stretches of x to
    # * convolve.
    # */

    _in = 0

    # ifdef _OPENMP
    # pragma omp parallel for firstprivate(in) schedule(static,1)
    # endif
    for _out in range(N):
        # /* width of resolution window for x is w = 2 dx ^ 2. */
        limit = -dx[_out] * xp[0]
        xo = x[_out]

        # /* Line up the left edge of the convolution window */
        # /* It is probably forward from the current position, */
        # /* but if the next dx is a lot higher than the current */
        # /* dx or if the x are not sorted, then it may be before */
        # /* the current position. * /
        # /* FIXME verify that the convolution window is just right */
        while _in < Nin - 1 and xin[_in] < xo - limit:
            _in += 1
        while _in > 0 and xin[_in] > xo - limit:
            _in -= 1

        # /* Special handling to avoid 0/0 for w = 0. */
        if dx[_out] > 0.0:
            # printf("convolve in:%ld out:%ld, xo:%g dx:%g\n",in,out,xo,dx[out])
            y[_out] = convolve_point_sampled(Nin, xin, yin, Np, xp, yp, xo, dx[_out], _in)
        elif _in < Nin - 1:
            # /* Linear interpolation */
            m = (yin[_in + 1] - yin[_in]) / (xin[_in + 1] - xin[_in])
            b = yin[_in] - m * xin[_in]
            y[_out] = m * xo + b
        elif _in > 0:
            # /* Linear extrapolation */
            m = (yin[_in] - yin[_in - 1]) / (xin[_in] - xin[_in - 1])
            b = yin[_in] - m * xin[_in]
            y[_out] = m * xo + b
        else:
            # /* Can't happen because there is more than one point in xin. */
            if not Nin > 1:
                raise ValueError("should be more than one point in Nin")



import numpy as np
from math import erf

verf = np.vectorize(erf)


def build_profile(z, offset, roughness, contrast, initial_value, profiles):
    """
    Convert a step profile to a smooth profile.

    *z*          calculation points, shape = (NZ,)
    *offset*     offset for each interface, shape = (NI,)
    *roughness*  roughness of each interface, shape = (NI,)
    *contrast*   step value at each interface for each profile, shape = (NP * NI)
    *initial_value* starting value for each profile, shape = (NP)

    *profiles*   (output) results of calculation, shape = (NP * NZ, order="C")
    """
    # Number of z values:
    NZ = len(z)
    # Number of wavelength values:
    NP = initial_value.shape[0]
    # Number of interfaces:
    NI = len(offset)

    contrast_shaped = contrast.reshape((NP, NI))
    profiles_shaped = profiles.reshape((NP, NZ))  # view - updates affect profiles

    profiles_shaped += initial_value.reshape((NP, 1))
    for i in range(NI):
        offset_i = offset[i]
        sigma_i = roughness[i]
        contrast_i = contrast_shaped[:, i]
        blended = blend(z, sigma_i, offset_i)
        delta = contrast_i.copy().reshape((NP, 1)) * blended.reshape((1, NZ))

        # delta = contrast_i[:, None] * blended[None, :]
        profiles_shaped += delta

    return


SQRT1_2 = 1.0 / np.sqrt(2.0)


def blend(z, sigma, offset):
    """
    blend function

    Given a Gaussian roughness value, compute the portion of the neighboring
    profile you expect to find in the current profile at depth z.
    """
    if sigma <= 0.0:
        return 1.0 * (z >= offset)
    else:
        return 0.5 * verf(SQRT1_2 * (z - offset) / sigma) + 0.5



# Some global variables for interactor

pick_radius = 5
layer_hysteresis = 4

thickness_color = "black"
interface_color = "black"
disabled_color = "gray"
active_color = "red"
rho_color = "black"
rhoI_color = "green"
rhoM_color = "blue"
thetaM_color = "orange"
profile_color = "cyan"
title_color = "red"

vf_scale = 0.4



"""
Reflectometry thickness interactor.
"""

from .config import thickness_color, pick_radius
from .interactor import BaseInteractor
from .util import clip


class ThicknessInteractor(BaseInteractor):
    """
    Control the size of the layers.
    """

    def __init__(self, profile):
        BaseInteractor.__init__(self, profile)
        self.markers = []
        self.textmark = []
        self._show_labels = True
        self._show_boundaries = True

    def label_offsets(self):
        z = self.profile.boundary[1:-1]
        left = -20
        middle = [(a + b) / 2.0 for a, b in zip(z[:-1], z[1:])]
        right = z[-1] + 20
        return [left] + middle + [right]

    def reset_markers(self):
        """
        Reset all markers.
        """
        self.clear_markers()
        ax = self.profile.axes

        # Add bars
        style = dict(
            linewidth=1,
            linestyle="-",
            color=thickness_color,
            alpha=0.75,
            pickradius=pick_radius,
            visible=self._show_boundaries,
        )
        self.markers = [ax.axvline(x=z, **style) for z in self.profile.boundary[1:-1]]

        fittable = [self.profile.sample_layer(idx).thickness.fittable for idx, _ in enumerate(self.markers)]
        fittable[0] = False  # First interface is not fittable
        for f, m in zip(fittable, self.markers):
            if not f:
                m.set(linestyle="--", linewidth=1.25)
        self.connect_markers(m for f, m in zip(fittable, self.markers) if f)

        # Add labels
        offsets = self.label_offsets()
        labels = self.profile.sample_labels()
        style = dict(
            transform=self.profile.xcoords,
            ha="left",
            va="bottom",
            rotation=30,
            fontsize="small",
            visible=self._show_labels,
        )
        self.textmark = [ax.text(z, 1, s, **style) for z, s in zip(offsets, labels)]

    def clear_markers(self):
        """
        Remove interfaces and layer names from the graph.
        """
        BaseInteractor.clear_markers(self)
        for h in self.textmark:
            h.remove()
        self.textmark = []

    def update_markers(self):
        """
        Update the marker positions
        """
        label_offsets = self.label_offsets()

        for z, h in zip(self.profile.boundary[1:-1], self.markers):
            h.set_xdata([z, z])
        for z, h in zip(label_offsets, self.textmark):
            h.set_x(z)

    def save(self, ev):
        idx = self.markers.index(ev.artist)
        curr = self.profile.sample_layer(idx).thickness
        next = self.profile.sample_layer(idx + 1).thickness if idx < len(self.markers) - 1 else None
        if not curr.fittable:
            curr = None
        if next is not None and not next.fittable:
            next = None

        self._idx = idx
        self._curr = curr
        self._next = next
        if curr is not None:
            self._curr_saved = curr.value
        if next is not None:
            self._next_saved = next.value

    def restore(self, ev):
        self._curr.value = self._curr_saved
        if self._next is not None:
            self._next.value = self._next_saved

    def drag(self, event):
        """
        Process move to a new position, making sure that the move is allowed.
        """
        if self._curr is None:
            return

        prev_offset = self.profile.boundary[self._idx]
        next_offset = self.profile.boundary[self._idx + 2]

        # Limit the position according to parameter limits
        if event.shift and self._next is not None:
            # shifting interface between two layers
            lo_curr, hi_curr = self._curr.prior.limits
            lo_next, hi_next = self._next.prior.limits
            lo = min(lo_curr + prev_offset, next_offset - hi_next)
            hi = min(hi_curr + prev_offset, next_offset - lo_next)
        else:
            # resizing a layer
            lo_curr, hi_curr = self._curr.prior.limits
            lo = lo_curr + prev_offset
            hi = hi_curr + prev_offset
        x = clip(event.xdata, lo, hi)

        # Set the current and next value based on offsets.  We set the next
        # in case the user is shifting the interface between two layers
        # instead of resizing the current layer.  We set it even when not
        # shifting in case the previous action was to shift, and the current
        # action triggers a restore.  We set the next before the current
        # so that if the two parameters are tied, the current takes
        # precedence.
        if self._next is not None:
            if event.shift:
                self._next.value = next_offset - x
            else:
                self._next.value = self._next_saved
        self._curr.value = x - prev_offset

        # Adjust x_offset such that the layer boundary will stay under the
        # curesor even if previous layers grow or shrink as a result of being
        # tied to the current layer thickness.
        self.profile._find_layer_boundaries()
        self.profile.x_offset -= self.profile.boundary[self._idx] - prev_offset



import os
from pathlib import Path
import shutil
import subprocess
import tempfile
from typing import Optional

import numpy as np
from numpy import radians

B2SLD = 2.31604654e-6
GEPORE_SRC = "gepore.f"
GEPORE_ZEEMAN_SRC = "gepore_zeeman.f"


class GeporeRunner:
    binary_name = "gepore"
    binary_zeeman_name = "gepore_zeeman"

    def __init__(self):
        self.binary_path = Path(tempfile.mkdtemp())
        self.source_folder = Path(__file__).parent
        self.gepore_src_path = self.source_folder / GEPORE_SRC
        self.gepore_zeeman_src_path = self.source_folder / GEPORE_ZEEMAN_SRC
        self.gepore_binary = self.binary_path / self.binary_name
        self.gepore_zeeman_binary = self.binary_path / self.binary_zeeman_name
        self._compile_gepore(self.gepore_binary, self.gepore_src_path)
        self._compile_gepore(self.gepore_zeeman_binary, self.gepore_zeeman_src_path)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.binary_path)

    def _compile_gepore(self, target: Path, source: Path):
        status = os.system(f"gfortran -o {target} {source}")
        if status != 0:
            raise RuntimeError("Could not compile %r" % source)
        if not target.exists():
            raise RuntimeError("No gepore created in %r" % target)

    def run(self, layers, QS, DQ, NQ, EPS, H, zeeman_corrections=True, output_folder: Optional[Path] = None):
        start_path = Path.cwd()  # save the current path

        if zeeman_corrections:
            layers = add_H(layers, H, EPS=EPS)
            gepore = self.gepore_zeeman_binary
        else:
            gepore = self.gepore_binary
        # layers = add_H(layers, H, EPS-270, 0)
        depth, rho, rhoB, thetaB, phiB = list(zip(*layers))

        NL = len(rho) - 2
        NC = 1
        ROSUP = rho[-1] + rhoB[-1]
        ROSUM = rho[-1] - rhoB[-1]
        ROINP = rho[0] + rhoB[0]
        ROINM = rho[0] - rhoB[0]

        path = Path(tempfile.mkdtemp())
        gepore_local = path / "gepore"
        shutil.copy(gepore, gepore_local)
        header = path / "inpt.d"
        layers = path / "tro.d"
        rm_real = path / "rrem.d"
        rm_imag = path / "rimm.d"
        rp_real = path / "rrep.d"
        rp_imag = path / "rimp.d"
        output_files = (header, layers, rm_real, rm_imag, rp_real, rp_imag)

        with open(layers, "w") as fid:
            for T, BN, PN, THE, PHI in list(zip(depth, rho, rhoB, thetaB, phiB))[1:-1]:
                fid.write(f"{T} {1e-6 * BN} {1e-6 * PN} {radians(THE)} {radians(PHI)}\n")
                # fid.write("%f %e %e %f %f\n" % (T, 1e-6 * BN, 1e-6 * PN, radians(THE), radians(PHI)))

        for IP, IM in ((0.0, 1.0), (1.0, 0.0)):
            with open(header, "w") as fid:
                fid.write(
                    f"{NL} {NC} {QS} {DQ} {NQ} {radians(EPS)} ({IP},0.0) ({IM},0.0) {1e-6 * ROINP} {1e-6 * ROINM} {1e-6 * ROSUP} {1e-6 * ROSUM}\n"
                    # "%d %d %f %f %d %f (%f,0.0) (%f,0.0) %e %e %e %e\n"
                    # % (NL, NC, QS, DQ, NQ, radians(EPS), IP, IM, 1e-6 * ROINP, 1e-6 * ROINM, 1e-6 * ROSUP, 1e-6 * ROSUM)
                )

            os.chdir(path)
            result = subprocess.run(["./gepore"], capture_output=True, check=True)

            rp = np.loadtxt(rp_real).T[1] + 1j * np.loadtxt(rp_imag).T[1]
            rm = np.loadtxt(rm_real).T[1] + 1j * np.loadtxt(rm_imag).T[1]
            if IP > 0.5:
                Rpp, Rpm = rp, rm
                if output_folder is not None:
                    plus_files = output_folder / "gepore_plus"
                    plus_files.mkdir(exist_ok=True, parents=True)
                    for f in output_files:
                        shutil.copy(f, plus_files)
            else:
                Rmp, Rmm = rp, rm
                if output_folder is not None:
                    minus_files = output_folder / "gepore_minus"
                    minus_files.mkdir(exist_ok=True, parents=True)
                    for f in output_files:
                        shutil.copy(f, minus_files)

        # clean up
        os.chdir(start_path)
        shutil.rmtree(path)
        return Rpp, Rpm, Rmp, Rmm


def add_H(layers, H=0.0, EPS=270.0):
    """Take H (vector) as input and add H to 4piM:"""
    new_layers = []
    for layer in layers:
        thickness, sld_n, sld_m, theta_m, phi_m = layer
        # we read phi_m, but it must be zero so we don't use it.
        sld_m_x = sld_m * np.cos(theta_m * np.pi / 180.0)  # phi_m = 0
        sld_m_y = sld_m * np.sin(theta_m * np.pi / 180.0)  # phi_m = 0
        sld_m_z = 0.0  # by Maxwell's equations, H_demag = mz so we'll just cancel it here
        sld_h = B2SLD * 1.0e6 * H
        # this code was completely wrong except for the case AGUIDE=270
        sld_h_x = 0  # by definition, H is along the z,lab direction and x,lab = x,sam so Hx,sam must = 0
        sld_h_y = -sld_h * np.sin(EPS * np.pi / 180.0)
        sld_h_z = sld_h * np.cos(EPS * np.pi / 180.0)
        sld_b_x = sld_h_x + sld_m_x
        sld_b_y = sld_h_y + sld_m_y
        sld_b_z = sld_h_z + sld_m_z
        sld_b = np.sqrt((sld_b_z) ** 2 + (sld_b_y) ** 2 + (sld_b_x) ** 2)
        # this was wrong:
        # theta_b = np.arctan2(sld_b_y, sld_b_x)
        theta_b = np.arccos(sld_b_x / sld_b)
        # this didn't hurt anything but is also unneeded:
        # theta_b = np.mod(theta_b, 2.0*np.pi)
        # this wasn't even close to correct:
        # phi_b = np.arcsin(sld_b_z/sld_b)
        phi_b = np.arctan2(sld_b_z, sld_b_y)
        phi_b = np.mod(phi_b, 2.0 * np.pi)
        new_layer = [thickness, sld_n, sld_b, theta_b * 180.0 / np.pi, phi_b * 180.0 / np.pi]
        new_layers.append(new_layer)
    return new_layers
